from .constant import (
    BATTERY_KEY,
    BUTTON_PRESSED_KEY,
    CHIMES_KEY,
    CONNECTION_KEY,
    MODEL_WIRED_VIDEO_DOORBELL,
    MODEL_WIREFREE_VIDEO_DOORBELL,
    MOTION_DETECTED_KEY,
    SIGNAL_STR_KEY,
    SILENT_MODE_ACTIVE_KEY,
    SILENT_MODE_CALL_KEY,
    SILENT_MODE_KEY,
    SIREN_STATE_KEY,
)
from .device import ArloChildDevice


class ArloDoorBell(ArloChildDevice):
    def __init__(self, name, arlo, attrs):
        super().__init__(name, arlo, attrs)
        self._motion_time_job = None
        self._ding_time_job = None
        self._has_motion_detect = False
        self._chimes = {}

    def _motion_stopped(self):
        self._save_and_do_callbacks(MOTION_DETECTED_KEY, False)
        with self._lock:
            self._motion_time_job = None

    def _button_unpressed(self):
        self._save_and_do_callbacks(BUTTON_PRESSED_KEY, False)
        with self._lock:
            self._ding_time_job = None

    def _event_handler(self, resource, event):
        self._arlo.debug(self.name + " DOORBELL got one " + resource)

        # create fake motion/button press event...
        if resource == self.resource_id:
            props = event.get("properties", {})

            # Newer doorbells send a motionDetected True followed by False. If we
            # see this then turn off connectionState checking.
            if MOTION_DETECTED_KEY in props:
                self._arlo.debug(self.name + " has motion detection support")
                self._has_motion_detect = True

            # Older doorbells signal a connectionState as available when motion
            # is detected. We check the properties length to not confuse it
            # with a device update. There is no motion stopped event so set a
            # timer to turn off the motion detect.
            if len(props) == 1 and not self._has_motion_detect:
                if props.get(CONNECTION_KEY, "") == "available":
                    self._save_and_do_callbacks(MOTION_DETECTED_KEY, True)
                    with self._lock:
                        self._arlo.bg.cancel(self._motion_time_job)
                        self._motion_time_job = self._arlo.bg.run_in(
                            self._motion_stopped, self._arlo.cfg.db_motion_time
                        )

            # For button presses we only get a buttonPressed notification, not
            # a "no longer pressed" notification - set a timer to turn off the
            # press.
            if BUTTON_PRESSED_KEY in props:
                self._save_and_do_callbacks(BUTTON_PRESSED_KEY, True)
                with self._lock:
                    self._arlo.bg.cancel(self._ding_time_job)
                    self._ding_time_job = self._arlo.bg.run_in(
                        self._button_unpressed, self._arlo.cfg.db_ding_time
                    )

            # Save out chimes
            if CHIMES_KEY in props:
                self._chimes = props[CHIMES_KEY]

            # Pass silent mode notifications so we can track them in the "ding"
            # entity.
            silent_mode = props.get(SILENT_MODE_KEY, {})
            if silent_mode:
                self._save_and_do_callbacks(SILENT_MODE_KEY, silent_mode)

        # pass on to lower layer
        super()._event_handler(resource, event)

    @property
    def resource_type(self):
        return "doorbells"

    @property
    def is_video_doorbell(self):
        return self.model_id.startswith(
            MODEL_WIRED_VIDEO_DOORBELL
        ) or self.model_id.startswith(MODEL_WIREFREE_VIDEO_DOORBELL)

    def has_capability(self, cap):
        if cap in (BUTTON_PRESSED_KEY, SILENT_MODE_KEY):
            return True
        if cap in (MOTION_DETECTED_KEY, BATTERY_KEY, SIGNAL_STR_KEY):
            # video doorbell provides these as a camera type
            if not self.is_video_doorbell:
                return True
        if cap in (CONNECTION_KEY,):
            # If video door bell is its own base station then don't provide connectivity here.
            if self.is_video_doorbell and self.parent_id == self.device_id:
                return False
        if cap in (SIREN_STATE_KEY,):
            if self.is_video_doorbell:
                return True
        return super().has_capability(cap)

    def update_silent_mode(self):
        """Requests the latest silent mode settings.

        Queues a job that requests the info from Arlo.
        """
        self._arlo.be.notify(
            base=self.base_station,
            body={
                "action": "get",
                "resource": self.resource_id,
                "publishResponse": False,
            },
        )

    def _build_chimes(self, on_or_off):
        chimes = {"traditional": on_or_off}
        for chime in self._chimes:
            chimes[chime] = on_or_off
        return chimes

    def _silence(self, active, calls, chimes):

        # Build settings
        silence_settings = {
            SILENT_MODE_ACTIVE_KEY: active,
            SILENT_MODE_CALL_KEY: calls,
        }
        if chimes:
            silence_settings[CHIMES_KEY] = chimes

        # Build request
        properties = {SILENT_MODE_KEY: silence_settings}
        self._arlo.debug(self.name + " silence is " + str(properties))

        # Send out request.
        response = self._arlo.be.notify(
            base=self.base_station,
            body={
                "action": "set",
                "properties": properties,
                "publishResponse": True,
                "resource": self.resource_id,
            },
            wait_for="response",
        )

        # Not none means a 200 so we assume it works until told otherwise.
        if response is not None:
            self._arlo.bg.run(
                self._save_and_do_callbacks,
                attr=SILENT_MODE_KEY,
                value=silence_settings,
            )

    def silence_off(self):
        self._silence(False, False, {})

    def silence_on(self):
        self._silence(True, True, self._build_chimes(True))

    def silence_chimes(self):
        self._silence(True, False, self._build_chimes(True))

    def silence_calls(self):
        self._silence(True, True, self._build_chimes(False))

    @property
    def is_silenced(self):
        return self._load(SILENT_MODE_KEY, {}).get(SILENT_MODE_ACTIVE_KEY, False)

    @property
    def calls_are_silenced(self):
        return self._load(SILENT_MODE_KEY, {}).get(SILENT_MODE_CALL_KEY, False)

    @property
    def chimes_are_silenced(self):
        for on_or_off in self._load(SILENT_MODE_KEY, {}).get(CHIMES_KEY, {}).values():
            if on_or_off is True:
                return True
        return False

    @property
    def _siren_resource_id(self):
        return "siren/{}".format(self.device_id)

    @property
    def siren_state(self):
        return self._load(SIREN_STATE_KEY, "off")

    def siren_on(self, duration=300, volume=8):
        """Turn camera siren on.

        Does nothing if camera doesn't support sirens.

        :param duration: how long, in seconds, to sound for
        :param volume: how long, from 1 to 8, to sound
        """
        body = {
            "action": "set",
            "resource": self._siren_resource_id,
            "publishResponse": True,
            "properties": {
                "sirenState": "on",
                "duration": int(duration),
                "volume": int(volume),
                "pattern": "alarm",
            },
        }
        self._arlo.be.notify(base=self, body=body)

    def siren_off(self):
        """Turn camera siren off.

        Does nothing if camera doesn't support sirens.
        """
        body = {
            "action": "set",
            "resource": self._siren_resource_id,
            "publishResponse": True,
            "properties": {"sirenState": "off"},
        }
        self._arlo.be.notify(base=self, body=body)



import pprint
import time

from .constant import (
    AIR_QUALITY_KEY,
    AUTOMATION_PATH,
    CONNECTION_KEY,
    DEFAULT_MODES,
    DEFINITIONS_PATH,
    HUMIDITY_KEY,
    MODE_ID_TO_NAME_KEY,
    MODE_IS_SCHEDULE_KEY,
    MODE_KEY,
    MODE_NAME_TO_ID_KEY,
    MODE_UPDATE_INTERVAL,
    MODEL_BABY,
    MODEL_WIREFREE_VIDEO_DOORBELL,
    MODEL_GO,
    PING_CAPABILITY,
    RESTART_PATH,
    SCHEDULE_KEY,
    SIREN_STATE_KEY,
    TEMPERATURE_KEY,
)
from .device import ArloDevice
from .util import time_to_arlotime

day_of_week = ["Mo", "Tu", "We", "Th", "Fr", "Sa", "Su", "Mo"]


class ArloBase(ArloDevice):
    def __init__(self, name, arlo, attrs):
        super().__init__(name, arlo, attrs)
        self._refresh_rate = 15
        self._schedules = None
        self._last_update = 0

    def _id_to_name(self, mode_id):
        return self._load([MODE_ID_TO_NAME_KEY, mode_id], None)

    def _id_is_schedule(self, mode_id):
        return self._load([MODE_IS_SCHEDULE_KEY, mode_id.lower()], False)

    def _name_to_id(self, mode_name):
        return self._load([MODE_NAME_TO_ID_KEY, mode_name.lower()], None)

    def _parse_modes(self, modes):
        for mode in modes:
            mode_id = mode.get("id", None)
            mode_name = mode.get("name", "")
            if mode_name == "":
                mode_name = mode.get("type", "")
                if mode_name == "":
                    mode_name = mode_id
            if mode_id and mode_name != "":
                self._arlo.debug(mode_id + "<=M=>" + mode_name)
                self._save([MODE_ID_TO_NAME_KEY, mode_id], mode_name)
                self._save([MODE_NAME_TO_ID_KEY, mode_name.lower()], mode_id)
                self._save([MODE_IS_SCHEDULE_KEY, mode_id.lower()], False)
                self._save([MODE_IS_SCHEDULE_KEY, mode_name.lower()], False)

    def schedule_to_modes(self):
        if self._schedules is None:
            return []

        now = time.localtime()
        day = day_of_week[now.tm_wday]
        minute = (now.tm_hour * 60) + now.tm_min
        for schedule in self._schedules:
            if not schedule.get("enabled", False):
                continue
            for action in schedule.get("schedule", []):
                if day in action.get("days", []):
                    start = action.get("startTime", 65535)
                    duration = action.get("duration", 65536)
                    if start <= minute < (start + duration):
                        modes = action.get("startActions", {}).get("enableModes", None)
                        if modes:
                            self._arlo.debug("schdule={}".format(modes[0]))
                            return modes

        # If nothing in schedule we are disarmed.
        return ['mode0']

    def _parse_schedules(self, schedules):
        self._schedules = schedules
        for schedule in schedules:
            schedule_id = schedule.get("id", None)
            schedule_name = schedule.get("name", "")
            if schedule_name == "":
                schedule_name = schedule_id
            if schedule_id and schedule_name != "":
                self._arlo.debug(schedule_id + "<=S=>" + schedule_name)
                self._save([MODE_ID_TO_NAME_KEY, schedule_id], schedule_name)
                self._save([MODE_NAME_TO_ID_KEY, schedule_name.lower()], schedule_id)
                self._save([MODE_IS_SCHEDULE_KEY, schedule_id.lower()], True)
                self._save([MODE_IS_SCHEDULE_KEY, schedule_name.lower()], True)

    def _set_mode_or_schedule(self, event):
        # schedule on or off?
        schedule_ids = event.get("activeSchedules", [])
        if schedule_ids:
            self._arlo.debug(self.name + " schedule change " + schedule_ids[0])
            schedule_name = self._id_to_name(schedule_ids[0])
            self._save_and_do_callbacks(SCHEDULE_KEY, schedule_name)
        else:
            self._arlo.debug(self.name + " schedule cleared ")
            self._save_and_do_callbacks(SCHEDULE_KEY, None)

        # mode present? we just set to that one... If no mode but schedule then
        # try to parse that out
        mode_ids = event.get("activeModes", [])
        if not mode_ids and schedule_ids:
            self._arlo.debug(self.name + " mode change (via schedule) ")
            self._arlo.vdebug(self.name + " schedules: " + pprint.pformat(self._schedules))
            mode_ids = self.schedule_to_modes()
        if mode_ids:
            self._arlo.debug(self.name + " mode change " + mode_ids[0])
            mode_name = self._id_to_name(mode_ids[0])
            self._save_and_do_callbacks(MODE_KEY, mode_name)

    def _event_handler(self, resource, event):
        self._arlo.debug(self.name + " BASE got " + resource)

        # modes on base station
        if resource == "modes":
            props = event.get("properties", {})

            # list of modes - recheck?
            self._parse_modes(props.get("modes", []))

            # mode change?
            if "activeMode" in props:
                self._save_and_do_callbacks(
                    MODE_KEY, self._id_to_name(props["activeMode"])
                )
            elif "active" in props:
                self._save_and_do_callbacks(MODE_KEY, self._id_to_name(props["active"]))

        # Base station mode change.
        # These come in per device and can arrive multiple times per state
        # change. We limit the updates to once per MODE_UPDATE_INTERVAL
        # seconds. Arlo doesn't send a "schedule changed" notification so we
        # re-fetch that information before testing the mode.
        elif resource == "states":
            now = time.monotonic()
            with self._lock:
                if now < self._last_update + MODE_UPDATE_INTERVAL:
                    return
                self._last_update = now
            self._arlo.debug("state change")
            self.update_modes()
            self.update_mode()

        # mode change?
        elif resource == "activeAutomations":
            self._set_mode_or_schedule(event)

        # schedule has changed, so reload
        elif resource == "automationRevisionUpdate":
            self.update_modes()

        # pass on to lower layer
        else:
            super()._event_handler(resource, event)

    @property
    def _v1_modes(self):
        if self._arlo.cfg.mode_api.lower() == "v1":
            self._arlo.vdebug("forced v1 api")
            return True
        if self._arlo.cfg.mode_api.lower() == "v2":
            self._arlo.vdebug("forced v2 api")
            return False
        if (
            self.model_id == MODEL_BABY
            or self.model_id == MODEL_GO
            or self.device_type == "arloq"
            or self.device_type == "arloqs"
        ):
            self._arlo.vdebug("deduced v1 api")
            return True
        else:
            self._arlo.vdebug("deduced v2 api")
            return False

    @property
    def available_modes(self):
        """Returns string list of available modes.

        For example:: ``['disarmed', 'armed', 'home']``
        """
        return list(self.available_modes_with_ids.keys())

    @property
    def available_modes_with_ids(self):
        """Returns dictionary of available modes mapped to Arlo ids.

        For example:: ``{'armed': 'mode1','disarmed': 'mode0','home': 'mode2'}``
        """
        modes = {}
        for key, mode_id in self._load_matching([MODE_NAME_TO_ID_KEY, "*"]):
            modes[key.split("/")[-1]] = mode_id
        if not modes:
            modes = DEFAULT_MODES
        return modes

    @property
    def mode(self):
        """Returns the current mode."""
        return self._load(MODE_KEY, "unknown")

    @mode.setter
    def mode(self, mode_name):
        """Set the base station mode.

        **Note:** Setting mode has been known to hang, method includes code to keep retrying.

        :param mode_name: mode to use, as returned by available_modes:
        """
        # Actually passed a mode?
        mode_id = None
        real_mode_name = self._id_to_name(mode_name)
        if real_mode_name:
            self._arlo.debug(f"passed an ID({mode_name}), converting it")
            mode_id = mode_name
            mode_name = real_mode_name

        # Need to change?
        if self.mode == mode_name:
            self._arlo.debug("no mode change needed")
            return

        if mode_id is None:
            mode_id = self._name_to_id(mode_name)
        if mode_id:

            # Need to change?
            if self.mode == mode_id:
                self._arlo.debug("no mode change needed (id)")
                return

            # Schedule or mode? Manually set schedule key.
            if self._id_is_schedule(mode_id):
                active = "activeSchedules"
                inactive = "activeModes"
                self._save_and_do_callbacks(SCHEDULE_KEY, mode_name)
            else:
                active = "activeModes"
                inactive = "activeSchedules"
                self._save_and_do_callbacks(SCHEDULE_KEY, None)

            # Post change.
            self._arlo.debug(self.name + ":new-mode=" + mode_name + ",id=" + mode_id)
            if self._v1_modes:
                self._arlo.be.notify(
                    base=self,
                    body={
                        "action": "set",
                        "resource": "modes",
                        "publishResponse": True,
                        "properties": {"active": mode_id},
                    },
                )
            else:
                # This is complicated... Setting a mode can fail and setting a mode can be sync or async.
                # This code tried 3 times to set the mode with attempts to reload the devices between
                # attempts to try and kick Arlo. In async mode the first set works in the current thread,
                # subsequent ones run in the background. In sync mode it the same. Sorry.
                def _set_mode_v2_cb(attempt):
                    self._arlo.debug("v2 arming")
                    params = {
                        "activeAutomations": [
                            {
                                "deviceId": self.device_id,
                                "timestamp": time_to_arlotime(),
                                active: [mode_id],
                                inactive: [],
                            }
                        ]
                    }
                    if attempt < 4:
                        tid = "(modes:{}|activeAutomations)".format(self.device_id)
                        body = self._arlo.be.post(
                            AUTOMATION_PATH,
                            params=params,
                            raw=True,
                            tid=tid,
                            wait_for=None,
                        )
                        if body is not None:
                            if (
                                body.get("success", False) is True
                                or body.get("resource", "") == "modes"
                                or body.get("resource", "") == "activeAutomations"
                            ):
                                return
                        self._arlo.warning(
                            "attempt {0}: error in response when setting mode=\n{1}".format(
                                attempt, pprint.pformat(body)
                            )
                        )
                        self._arlo.debug(
                            "Fetching device list (hoping this will fix arming/disarming)"
                        )
                        self._arlo.be.devices()
                        if self._arlo.cfg.synchronous_mode:
                            self._arlo.debug("trying again, but synchronous")
                            _set_mode_v2_cb(attempt=attempt + 1)
                        else:
                            self._arlo.bg.run(_set_mode_v2_cb, attempt=attempt + 1)
                        return

                    self._arlo.error("Failed to set mode.")
                    self._arlo.debug(
                        "Giving up on setting mode! Session headers=\n{}".format(
                            pprint.pformat(self._arlo.be.session.headers)
                        )
                    )
                    self._arlo.debug(
                        "Giving up on setting mode! Session cookies=\n{}".format(
                            pprint.pprint(self._arlo.be.session.cookies)
                        )
                    )

                _set_mode_v2_cb(1)
        else:
            self._arlo.warning(
                "{0}: mode {1} is unrecognised".format(self.name, mode_name)
            )

    def update_mode(self):
        """Check and update the base's current mode."""
        now = time.monotonic()
        with self._lock:
            #  if now < self._last_update + MODE_UPDATE_INTERVAL:
            #  self._arlo.debug('skipping an update')
            #  return
            self._last_update = now
        data = self._arlo.be.get(AUTOMATION_PATH)
        for mode in data:
            if mode.get("uniqueId", "") == self.unique_id:
                self._set_mode_or_schedule(mode)

    def update_modes(self, initial=False):
        """Get and update the available modes for the base."""
        if self._v1_modes:
            # Work around slow arlo connections.
            if initial and self._arlo.cfg.synchronous_mode:
                time.sleep(5)
            resp = self._arlo.be.notify(
                base=self,
                body={"action": "get", "resource": "modes", "publishResponse": False},
                wait_for="event",
            )
            if resp is not None:
                props = resp.get("properties", {})
                self._parse_modes(props.get("modes", []))
            else:
                self._arlo.error("unable to read mode, try forcing v2")
        else:
            modes = self._arlo.be.get(
                DEFINITIONS_PATH + "?uniqueIds={}".format(self.unique_id)
            )
            if modes is not None:
                modes = modes.get(self.unique_id, {})
                self._parse_modes(modes.get("modes", []))
                self._parse_schedules(modes.get("schedules", []))
            else:
                self._arlo.error("failed to read modes (v2)")

    @property
    def schedule(self):
        """Returns current schedule name or `None` if no schedule active."""
        return self._load(SCHEDULE_KEY, None)

    @property
    def on_schedule(self):
        """Returns `True` is base station is running a schedule."""
        return self.schedule is not None

    @property
    def refresh_rate(self):
        return self._refresh_rate

    @refresh_rate.setter
    def refresh_rate(self, value):
        if isinstance(value, (int, float)):
            self._refresh_rate = value

    @property
    def siren_state(self):
        """Returns the current siren state (`on` or `off`)."""
        return self._load(SIREN_STATE_KEY, "off")

    def siren_on(self, duration=300, volume=8):
        """Turn base siren on.

        Does nothing if base doesn't support sirens.

        :param duration: how long, in seconds, to sound for
        :param volume: how long, from 1 to 8, to sound
        """
        body = {
            "action": "set",
            "resource": "siren",
            "publishResponse": True,
            "properties": {
                "sirenState": "on",
                "duration": int(duration),
                "volume": int(volume),
                "pattern": "alarm",
            },
        }
        self._arlo.debug(str(body))
        self._arlo.be.notify(base=self, body=body)

    def siren_off(self):
        """Turn base siren off.

        Does nothing if base doesn't support sirens.
        """
        body = {
            "action": "set",
            "resource": "siren",
            "publishResponse": True,
            "properties": {"sirenState": "off"},
        }
        self._arlo.debug(str(body))
        self._arlo.be.notify(base=self, body=body)

    def restart(self):
        params = {"deviceId": self.device_id}
        tid = "diagnostics:{}".format(self.device_id)
        if (
            self._arlo.be.post(RESTART_PATH, params=params, tid=tid, wait_for=None)
            is None
        ):
            self._arlo.debug("RESTART didnt send")

    def _ping_and_check_reply(self):
        body = {
            "action": "set",
            "resource": self._arlo.be.sub_id,
            "publishResponse": False,
            "properties": {"devices": [self.device_id]},
        }
        self._arlo.debug("pinging {}".format(self.name))
        if self._arlo.be.notify(base=self, body=body, wait_for="response") is None:
            self._save_and_do_callbacks(CONNECTION_KEY, "unavailable")
        else:
            self._save_and_do_callbacks(CONNECTION_KEY, "available")

    def ping(self):
        self._arlo.bg.run(self._ping_and_check_reply)

    @property
    def state(self):
        if self.is_unavailable:
            return "unavailable"
        return "available"

    def has_capability(self, cap):
        if cap in (TEMPERATURE_KEY, HUMIDITY_KEY, AIR_QUALITY_KEY):
            if self.model_id.startswith(MODEL_BABY):
                return True
        if cap in (SIREN_STATE_KEY,):
            if (
                self.model_id.startswith(("VMB400", "VMB450"))
                or self.model_id == MODEL_GO
            ):
                return True
        if cap in (PING_CAPABILITY,):
            # Battery powered wifi devices that act as their own base station don't get pinged.
            if self.model_id.startswith(MODEL_BABY):
                return True
            if self.is_own_parent and self.using_wifi and not self.is_corded:
                return False
            # Wire free video doorbell acting as base station
            if (
                self.is_own_parent
                and self.model_id.startswith(MODEL_WIREFREE_VIDEO_DOORBELL)
                and not self.is_corded
            ):
                return False
            return True
        return super().has_capability(cap)



import json
import pprint
import re
import threading
import time
import traceback
import uuid

import cloudscraper
import requests
import requests.adapters

from .constant import (
    AUTH_FINISH_PATH,
    AUTH_GET_FACTORS,
    AUTH_PATH,
    AUTH_START_PATH,
    AUTH_VALIDATE_PATH,
    DEFAULT_RESOURCES,
    DEVICES_PATH,
    LOGOUT_PATH,
    NOTIFY_PATH,
    SESSION_PATH,
    SUBSCRIBE_PATH,
    TFA_CONSOLE_SOURCE,
    TFA_IMAP_SOURCE,
    TFA_PUSH_SOURCE,
    TFA_REST_API_SOURCE,
    TRANSID_PREFIX,
)
from .sseclient import SSEClient
from .tfa import Arlo2FAConsole, Arlo2FAImap, Arlo2FARestAPI
from .util import now_strftime, time_to_arlotime, to_b64


# include token and session details
class ArloBackEnd(object):
    def __init__(self, arlo):

        self._arlo = arlo
        self._lock = threading.Condition()
        self._req_lock = threading.Lock()

        self._dump_file = self._arlo.cfg.dump_file

        self._requests = {}
        self._callbacks = {}
        self._resource_types = DEFAULT_RESOURCES

        self._token = None
        self._user_id = None
        self._web_id = None
        self._sub_id = None

        self._ev_stream = None

        # login
        self._session = None
        self._logged_in = self._login()
        if not self._logged_in:
            self._arlo.debug("failed to log in")
            return

        # event loop thread - started as needed
        self._ev_start()

        # start logout daemon
        if self._arlo.cfg.reconnect_every != 0:
            self._arlo.debug("automatically reconnecting")
            self._arlo.bg.run_every(self.logout, self._arlo.cfg.reconnect_every)

    def _request(
        self,
        path,
        method="GET",
        params=None,
        headers=None,
        stream=False,
        raw=False,
        timeout=None,
        host=None,
    ):
        if params is None:
            params = {}
        if headers is None:
            headers = {}
        if timeout is None:
            timeout = self._arlo.cfg.request_timeout
        try:
            with self._req_lock:
                if host is None:
                    host = self._arlo.cfg.host
                url = host + path
                self._arlo.vdebug("request-url={}".format(url))
                self._arlo.vdebug("request-params=\n{}".format(pprint.pformat(params)))
                self._arlo.vdebug(
                    "request-headers=\n{}".format(pprint.pformat(headers))
                )
                if method == "GET":
                    r = self._session.get(
                        url,
                        params=params,
                        headers=headers,
                        stream=stream,
                        timeout=timeout,
                    )
                    if stream is True:
                        return r
                elif method == "PUT":
                    r = self._session.put(
                        url, json=params, headers=headers, timeout=timeout
                    )
                elif method == "POST":
                    r = self._session.post(
                        url, json=params, headers=headers, timeout=timeout
                    )
        except Exception as e:
            self._arlo.warning("request-error={}".format(type(e).__name__))
            return None

        self._arlo.vdebug("request-end={}".format(r.status_code))
        if r.status_code != 200:
            return None

        try:
            body = r.json()
            self._arlo.vdebug("request-body=\n{}".format(pprint.pformat(body)))
        except Exception as e:
            self._arlo.warning("body-error={}".format(type(e).__name__))
            return None

        if raw:
            return body

        # New auth style and TFA helper
        if "meta" in body:
            if body["meta"]["code"] == 200:
                return body["data"]
            else:
                self._arlo.warning("error in new response=" + str(body))

        # Original response type
        elif "success" in body:
            if body["success"]:
                if "data" in body:
                    return body["data"]
                # success, but no data so fake empty data
                return {}
            else:
                self._arlo.warning("error in response=" + str(body))

        return None

    def gen_trans_id(self, trans_type=TRANSID_PREFIX):
        return trans_type + "!" + str(uuid.uuid4())

    def _ev_reconnected(self):
        self._arlo.debug("Fetching device list after ev-reconnect")
        self.devices()

    def _ev_dispatcher(self, response):

        # get message type(s) and id(s)
        responses = []
        resource = response.get("resource", "")

        err = response.get("error", None)
        if err is not None:
            self._arlo.info(
                "error: code="
                + str(err.get("code", "xxx"))
                + ",message="
                + str(err.get("message", "XXX"))
            )

        #
        # I'm trying to keep this as generic as possible... but it needs some
        # smarts to figure out where to send responses - the packets from Arlo
        # are anything but consistent...
        # See docs/packets for and idea of what we're parsing.
        #

        # Answer for async ping. Note and finish.
        # Packet number #1.
        if resource.startswith("subscriptions/"):
            self._arlo.vdebug("async ping response " + resource)
            return

        # These is a base station mode response. Find base station ID and
        # forward response.
        # Packet number #4.
        if resource == "activeAutomations":
            for device_id in response:
                if device_id != "resource":
                    responses.append((device_id, resource, response[device_id]))

        # Mode update response
        elif "states" in response:
            device_id = response.get("from", None)
            if device_id is not None:
                responses.append((device_id, "states", response["states"]))

        # These are individual device responses. Find device ID and forward
        # response.
        # Packet number #?.
        elif [x for x in self._resource_types if resource.startswith(x + "/")]:
            device_id = resource.split("/")[1]
            responses.append((device_id, resource, response))

        # These are base station responses. Which can be about the base station
        # or devices on it... Check if property is list.
        # Packet number #3/#2
        elif resource in self._resource_types:
            prop_or_props = response.get("properties", [])
            if isinstance(prop_or_props, list):
                for prop in prop_or_props:
                    device_id = prop.get("serialNumber", None)
                    if device_id is None:
                        device_id = response.get("from", None)
                    responses.append((device_id, resource, prop))
            else:
                device_id = response.get("from", None)
                responses.append((device_id, resource, response))

        elif resource.startswith("audioPlayback"):
            device_id = response.get("from")
            properties = response.get("properties")
            if resource == "audioPlayback/status":
                # Wrap the status event to match the 'audioPlayback' event
                properties = {"status": response.get("properties")}

            self._arlo.info(
                "audio playback response {} - {}".format(resource, response)
            )
            if device_id is not None and properties is not None:
                responses.append((device_id, resource, properties))

        # This a list ditch effort to funnel the answer the correct place...
        #  Check for device_id
        #  Check for unique_id
        # If none of those then is unhandled
        # Packet number #?.
        else:
            device_id = response.get("deviceId", None)
            if device_id is not None:
                responses.append((device_id, resource, response))
            else:
                device_id = response.get("uniqueId", None)
                if device_id is not None:
                    responses.append((device_id, resource, response))
                else:
                    self._arlo.debug(
                        "unhandled response {} - {}".format(resource, response)
                    )

        # Now find something waiting for this/these.
        for device_id, resource, response in responses:
            cbs = []
            self._arlo.debug("sending {} to {}".format(resource, device_id))
            with self._lock:
                if device_id and device_id in self._callbacks:
                    cbs.extend(self._callbacks[device_id])
                if "all" in self._callbacks:
                    cbs.extend(self._callbacks["all"])
            for cb in cbs:
                self._arlo.bg.run(cb, resource=resource, event=response)

    def _ev_loop(self, stream):

        # say we're starting
        if self._dump_file is not None:
            with open(self._dump_file, "a") as dump:
                time_stamp = now_strftime("%Y-%m-%d %H:%M:%S.%f")
                dump.write("{}: {}\n".format(time_stamp, "ev_loop start"))

        # for event in stream.events():
        for event in stream:

            # stopped?
            if event is None:
                with self._lock:
                    self._ev_connected_ = False
                    self._lock.notify_all()
                break

            # dig out response, print out verbose debug
            response = json.loads(event.data)
            if self._dump_file is not None:
                with open(self._dump_file, "a") as dump:
                    time_stamp = now_strftime("%Y-%m-%d %H:%M:%S.%f")
                    dump.write(
                        "{}: {}\n".format(
                            time_stamp, pprint.pformat(response, indent=2)
                        )
                    )
            self._arlo.vdebug(
                "packet-in=\n{}".format(pprint.pformat(response, indent=2))
            )

            # logged out? signal exited
            if response.get("action") == "logout":
                with self._lock:
                    self._ev_connected_ = False
                    self._requests = {}
                    self._lock.notify_all()
                self._arlo.warning("logged out? did you log in from elsewhere?")
                break

            # connected - yay!
            if response.get("status") == "connected":
                with self._lock:
                    self._ev_connected_ = True
                    self._lock.notify_all()
                continue

            # Run the dispatcher to set internal state and run callbacks.
            self._ev_dispatcher(response)

            # is there a notify/post waiting for this response? If so, signal to waiting entity.
            tid = response.get("transId", None)
            resource = response.get("resource", None)
            device_id = response.get("from", None)
            with self._lock:
                # Transaction ID
                # Simple. We have a transaction ID, look for that. These are
                # usually returned by notify requests.
                if tid and tid in self._requests:
                    self._requests[tid] = response
                    self._lock.notify_all()

                # Resource
                # These are usually returned after POST requests. We trap these
                # to make async calls sync.
                if resource:
                    # Historical. We are looking for a straight matching resource.
                    if resource in self._requests:
                        self._arlo.vdebug("{} found by text!".format(resource))
                        self._requests[resource] = response
                        self._lock.notify_all()
                    else:
                        # Complex. We are looking for a resource and-or
                        # deviceid matching a regex.
                        if device_id:
                            resource = "{}:{}".format(resource, device_id)
                            self._arlo.vdebug("{} bounded device!".format(resource))
                        for request in self._requests:
                            if re.match(request, resource):
                                self._arlo.vdebug(
                                    "{} found by regex {}!".format(resource, request)
                                )
                                self._requests[request] = response
                                self._lock.notify_all()

    def _ev_thread_main(self):

        self._arlo.debug("starting event loop")
        while True:

            # login again if not first iteration, this will also create a new session
            while not self._logged_in:
                with self._lock:
                    self._lock.wait(5)
                self._arlo.debug("re-logging in")
                self._logged_in = self._login()

            # get stream, restart after requested seconds of inactivity or forced close
            try:
                if self._arlo.cfg.stream_timeout == 0:
                    self._arlo.debug("starting stream with no timeout")
                    # self._ev_stream = SSEClient( self.get( SUBSCRIBE_PATH + self._token,stream=True,raw=True ) )
                    self._ev_stream = SSEClient(
                        self._arlo,
                        self._arlo.cfg.host + SUBSCRIBE_PATH,
                        session=self._session,
                        reconnect_cb=self._ev_reconnected,
                    )
                else:
                    self._arlo.debug(
                        "starting stream with {} timeout".format(
                            self._arlo.cfg.stream_timeout
                        )
                    )
                    # self._ev_stream = SSEClient(
                    #     self.get(SUBSCRIBE_PATH + self._token, stream=True, raw=True,
                    #              timeout=self._arlo.cfg.stream_timeout))
                    self._ev_stream = SSEClient(
                        self._arlo,
                        self._arlo.cfg.host + SUBSCRIBE_PATH,
                        session=self._session,
                        reconnect_cb=self._ev_reconnected,
                        timeout=self._arlo.cfg.stream_timeout,
                    )
                self._ev_loop(self._ev_stream)
            except requests.exceptions.ConnectionError:
                self._arlo.warning("event loop timeout")
            except AttributeError as e:
                self._arlo.warning("forced close " + str(e))
            except Exception as e:
                # self._arlo.warning('general exception ' + str(e))
                self._arlo.error(
                    "general-error={}\n{}".format(
                        type(e).__name__, traceback.format_exc()
                    )
                )

            # restart login...
            self._ev_stream = None
            self._logged_in = False

    def _ev_start(self):
        self._ev_stream = None
        self._ev_connected_ = False
        self._ev_thread = threading.Thread(
            name="ArloEventStream", target=self._ev_thread_main, args=()
        )
        self._ev_thread.setDaemon(True)

        with self._lock:
            self._ev_thread.start()
            if not self._ev_connected_:
                self._arlo.debug("waiting for stream up")
                self._lock.wait(30)

        self._arlo.debug("stream up")
        return True

    def _get_tfa(self):
        """Return the 2FA type we're using."""
        tfa_type = self._arlo.cfg.tfa_source
        if tfa_type == TFA_CONSOLE_SOURCE:
            return Arlo2FAConsole(self._arlo)
        elif tfa_type == TFA_IMAP_SOURCE:
            return Arlo2FAImap(self._arlo)
        elif tfa_type == TFA_REST_API_SOURCE:
            return Arlo2FARestAPI(self._arlo)
        else:
            return tfa_type

    def _update_auth_info(self, body):
        self._token = body["token"]
        self._token64 = to_b64(self._token)
        self._user_id = body["userId"]
        self._web_id = self._user_id + "_web"
        self._sub_id = "subscriptions/" + self._web_id

    def _auth(self):
        headers = {
            "Auth-Version": "2",
            "Accept": "application/json, text/plain, */*",
            "Referer": self._arlo.cfg.host,
            "User-Agent": self._user_agent,
            "Source": "arloCamWeb",
        }

        # Handle 1015 error
        attempt = 0
        body = None
        while attempt < 3:
            attempt += 1
            self._arlo.debug("login attempt #{}".format(attempt))
            body = self.auth_post(
                AUTH_PATH,
                {
                    "email": self._arlo.cfg.username,
                    "password": to_b64(self._arlo.cfg.password),
                    "language": "en",
                    "EnvSource": "prod",
                },
                headers,
            )
            if body is not None:
                break
            time.sleep(1)
        if body is None:
            self._arlo.error("authentication failed")
            return False

        # save new login information
        self._update_auth_info(body)

        # Looks like we need 2FA. So, request a code be sent to our email address.
        if not body["authCompleted"]:
            self._arlo.debug("need 2FA...")

            # update headers and create 2fa instance
            headers["Authorization"] = self._token64
            tfa = self._get_tfa()

            # get available 2fa choices,
            self._arlo.debug("getting tfa choices")
            factors = self.auth_get(
                AUTH_GET_FACTORS + "?data = {}".format(int(time.time())), {}, headers
            )
            if factors is None:
                self._arlo.error("2fa: no secondary choices available")
                return False

            # look for code source choice
            self._arlo.debug("looking for {}".format(self._arlo.cfg.tfa_type))
            factor_id = None
            for factor in factors["items"]:
                if factor["factorType"].lower() == self._arlo.cfg.tfa_type:
                    factor_id = factor["factorId"]
            if factor_id is None:
                self._arlo.error("2fa no suitable secondary choice available")
                return False

            if tfa != TFA_PUSH_SOURCE:
                # snapshot 2fa before sending in request
                if not tfa.start():
                    self._arlo.error("2fa startup failed")
                    return False

                # start authentication with email
                self._arlo.debug(
                    "starting auth with {}".format(self._arlo.cfg.tfa_type)
                )
                body = self.auth_post(AUTH_START_PATH, {"factorId": factor_id}, headers)
                if body is None:
                    self._arlo.error("2fa startAuth failed")
                    return False
                factor_auth_code = body["factorAuthCode"]

                # get code from TFA source
                code = tfa.get()
                if code is None:
                    self._arlo.error("2fa core retrieval failed")
                    return False

                # tidy 2fa
                tfa.stop()

                # finish authentication
                self._arlo.debug("finishing auth")
                body = self.auth_post(
                    AUTH_FINISH_PATH,
                    {"factorAuthCode": factor_auth_code, "otp": code},
                    headers,
                )
                if body is None:
                    self._arlo.error("2fa finishAuth failed")
                    return False
            else:
                # start authentication
                self._arlo.debug(
                    "starting auth with {}".format(self._arlo.cfg.tfa_type)
                )
                body = self.auth_post(AUTH_START_PATH, {"factorId": factor_id}, headers)
                if body is None:
                    self._arlo.error("2fa startAuth failed")
                    return False
                factor_auth_code = body["factorAuthCode"]
                tries = 1
                while True:
                    # finish authentication
                    self._arlo.debug("finishing auth")
                    body = self.auth_post(
                        AUTH_FINISH_PATH,
                        {"factorAuthCode": factor_auth_code},
                        headers,
                    )
                    if body is None:
                        self._arlo.warning("2fa finishAuth - tries {}".format(tries))
                        if tries < self._arlo.cfg.tfa_retries:
                            time.sleep(self._arlo.cfg.tfa_delay)
                            tries += 1
                        else:
                            self._arlo.error("2fa finishAuth failed")
                            return False
                    else:
                        break

            # save new login information
            self._update_auth_info(body)

        return True

    def _validate(self):
        headers = {
            "Auth-Version": "2",
            "Accept": "application/json, text/plain, */*",
            "Authorization": self._token64,
            "Referer": self._arlo.cfg.host,
            "User-Agent": self._user_agent,
            "Source": "arloCamWeb",
        }

        # Validate it!
        validated = self.auth_get(
            AUTH_VALIDATE_PATH + "?data = {}".format(int(time.time())), {}, headers
        )
        if validated is None:
            self._arlo.error("token validation failed")
            return False
        return True

    def _v2_session(self):
        v2_session = self.get(SESSION_PATH)
        if v2_session is None:
            self._arlo.error("session start failed")
            return False
        return True

    def _login(self):

        # pickup user configured user agent
        self._user_agent = self.user_agent(self._arlo.cfg.user_agent)

        # set up session
        self._session = cloudscraper.create_scraper()
        #  if self._arlo.cfg.http_connections != 0 and self._arlo.cfg.http_max_size != 0:
        #  self._arlo.debug(
        #  "custom connections {}:{}".format(
        #  self._arlo.cfg.http_connections, self._arlo.cfg.http_max_size
        #  )
        #  )
        #  self._session.mount(
        #  "https://",
        #  requests.adapters.HTTPAdapter(
        #  pool_connections=self._arlo.cfg.http_connections,
        #  pool_maxsize=self._arlo.cfg.http_max_size,
        #  ),
        #  )

        if not self._auth():
            return False

        if not self._validate():
            return False

        # update sessions headers
        headers = {
            "Accept": "application/json, text/plain, */*",
            "Accept-Language": "en-GB,en;q=0.9,en-US;q=0.8",
            "Auth-Version": "2",
            "Cache-Control": "no-cache",
            "SchemaVersion": "1",
            "Host": re.sub("https?://", "", self._arlo.cfg.host),
            "Content-Type": "application/json; charset=utf-8;",
            "Origin": self._arlo.cfg.host,
            "Pragma": "no-cache",
            "Referer": self._arlo.cfg.host,
            "User-Agent": self._user_agent,
            "Authorization": self._token,
        }
        self._session.headers.update(headers)

        if not self._v2_session():
            return False
        return True

    def _notify(self, base, body, trans_id=None):
        if trans_id is None:
            trans_id = self.gen_trans_id()

        body["to"] = base.device_id
        body["from"] = self._web_id
        body["transId"] = trans_id

        if (
            self.post(
                NOTIFY_PATH + base.device_id, body, headers={"xcloudId": base.xcloud_id}
            )
            is None
        ):
            return None
        return trans_id

    def _start_transaction(self, tid=None):
        if tid is None:
            tid = self.gen_trans_id()
        self._arlo.vdebug("starting transaction-->{}".format(tid))
        with self._lock:
            self._requests[tid] = None
        return tid

    def _wait_for_transaction(self, tid, timeout):
        if timeout is None:
            timeout = self._arlo.cfg.request_timeout
        mnow = time.monotonic()
        mend = mnow + timeout

        self._arlo.vdebug("finishing transaction-->{}".format(tid))
        with self._lock:
            try:
                while mnow < mend and self._requests[tid] is None:
                    self._lock.wait(mend - mnow)
                    mnow = time.monotonic()
                response = self._requests.pop(tid)
            except KeyError as _e:
                self._arlo.debug("got a key error")
                response = None
        self._arlo.vdebug("finished transaction-->{}".format(tid))
        return response

    @property
    def is_connected(self):
        return self._logged_in

    def logout(self):
        self._arlo.debug("trying to logout")
        if self._ev_stream is not None:
            self._ev_stream.stop()
        self.put(LOGOUT_PATH)

    def notify(self, base, body, timeout=None, wait_for=None):
        """Send in a notification.

        Notifications are Arlo's way of getting stuff done - turn on a light, change base station mode,
        start recording. Pyaarlo will post a notification and Arlo will post a reply on the event
        stream indicating if it worked or not or of a state change.

        How Pyaarlo treats notifications depends on the mode it's being run in. For asynchronous mode - the
        default - it sends the notification and returns immediately. For synchronous mode it sends the
        notification and waits for the event related to the notification to come back. To use the default
        settings leave `wait_for` as `None`, to force asynchronous set `wait_for` to `nothing` and to force
        synchronous set `wait_for` to `event`.

        There is a third way to send a notification where the code waits for the initial response to come back
        but that must be specified by setting `wait_for` to `response`.

        :param base: base station to use
        :param body: notification message
        :param timeout: how long to wait for response before failing, only applied if `wait_for` is `event`.
        :param wait_for: what to wait for, either `None`, `event`, `response` or `nothing`.
        :return: either a response packet or an event packet
        """
        if wait_for is None:
            wait_for = "event" if self._arlo.cfg.synchronous_mode else "nothing"

        if wait_for == "event":
            self._arlo.vdebug("notify+event running")
            tid = self._start_transaction()
            self._notify(base, body=body, trans_id=tid)
            return self._wait_for_transaction(tid, timeout)
            # return self._notify_and_get_event(base, body, timeout=timeout)
        elif wait_for == "response":
            self._arlo.vdebug("notify+response running")
            return self._notify(base, body=body)
        else:
            self._arlo.vdebug("notify+ sent")
            self._arlo.bg.run(self._notify, base=base, body=body)

    def get(
        self,
        path,
        params=None,
        headers=None,
        stream=False,
        raw=False,
        timeout=None,
        host=None,
        wait_for="response",
    ):
        if wait_for == "response":
            self._arlo.vdebug("get+response running")
            return self._request(
                path, "GET", params, headers, stream, raw, timeout, host
            )
        else:
            self._arlo.vdebug("get sent")
            self._arlo.bg.run(
                self._request, path, "GET", params, headers, stream, raw, timeout, host
            )

    def put(
        self,
        path,
        params=None,
        headers=None,
        raw=False,
        timeout=None,
        wait_for="response",
    ):
        if wait_for == "response":
            self._arlo.vdebug("put+response running")
            return self._request(path, "PUT", params, headers, False, raw, timeout)
        else:
            self._arlo.vdebug("put sent")
            self._arlo.bg.run(
                self._request, path, "PUT", params, headers, False, raw, timeout
            )

    def post(
        self,
        path,
        params=None,
        headers=None,
        raw=False,
        timeout=None,
        tid=None,
        wait_for="response",
    ):
        """Post a request to the Arlo servers.

        Posts are used to retrieve data from the Arlo servers. Mostly. They are also used to change
        base station modes.

        The default mode of operation is to wait for a response from the http request. The `wait_for`
        variable can change the operation. Setting it to `response` waits for a http response.
        Setting it to `resource` waits for the resource in the `params` parameter to appear in the event
        stream. Setting it to `nothing` causing the post to run in the background. Setting it to `None`
        uses `resource` in synchronous mode and `response` in asynchronous mode.
        """
        if wait_for is None:
            wait_for = "resource" if self._arlo.cfg.synchronous_mode else "response"

        if wait_for == "resource":
            self._arlo.vdebug("notify+resource running")
            if tid is None:
                tid = list(params.keys())[0]
            tid = self._start_transaction(tid)
            self._request(path, "POST", params, headers, False, raw, timeout)
            return self._wait_for_transaction(tid, timeout)
        if wait_for == "response":
            self._arlo.vdebug("post+response running")
            return self._request(path, "POST", params, headers, False, raw, timeout)
        else:
            self._arlo.vdebug("post sent")
            self._arlo.bg.run(
                self._request, path, "POST", params, headers, False, raw, timeout
            )

    def auth_post(self, path, params=None, headers=None, raw=False, timeout=None):
        return self._request(
            path, "POST", params, headers, False, raw, timeout, self._arlo.cfg.auth_host
        )

    def auth_get(
        self, path, params=None, headers=None, stream=False, raw=False, timeout=None
    ):
        return self._request(
            path, "GET", params, headers, stream, raw, timeout, self._arlo.cfg.auth_host
        )

    @property
    def session(self):
        return self._session

    @property
    def sub_id(self):
        return self._sub_id

    def add_listener(self, device, callback):
        with self._lock:
            if device.device_id not in self._callbacks:
                self._callbacks[device.device_id] = []
            self._callbacks[device.device_id].append(callback)
            if device.unique_id not in self._callbacks:
                self._callbacks[device.unique_id] = []
            self._callbacks[device.unique_id].append(callback)

    def add_any_listener(self, callback):
        with self._lock:
            if "all" not in self._callbacks:
                self._callbacks["all"] = []
            self._callbacks["all"].append(callback)

    def del_listener(self, device, callback):
        pass

    def devices(self):
        return self.get(DEVICES_PATH + "?t={}".format(time_to_arlotime()))

    def user_agent(self, agent):
        """Map `agent` to a real user agent.

        User provides a default user agent they want for most interactions but it can be overridden
        for stream operations.
        """
        self._arlo.debug(f"looking for user_agent {agent}")
        if agent.lower() == "arlo":
            return (
                "Mozilla/5.0 (iPhone; CPU iPhone OS 11_1_2 like Mac OS X) "
                "AppleWebKit/604.3.5 (KHTML, like Gecko) Mobile/15B202 NETGEAR/v1 "
                "(iOS Vuezone)"
            )
        elif agent.lower() == "iphone":
            return (
                "Mozilla/5.0 (iPhone; CPU iPhone OS 13_1_3 like Mac OS X) "
                "AppleWebKit/605.1.15 (KHTML, like Gecko) Version/13.0.1 Mobile/15E148 Safari/604.1"
            )
        elif agent.lower() == "ipad":
            return (
                "Mozilla/5.0 (iPad; CPU OS 12_2 like Mac OS X) "
                "AppleWebKit/605.1.15 (KHTML, like Gecko) Version/12.1 Mobile/15E148 Safari/604.1"
            )
        elif agent.lower() == "mac":
            return (
                "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_11_6) "
                "AppleWebKit/605.1.15 (KHTML, like Gecko) Version/11.1.2 Safari/605.1.15"
            )
        elif agent.lower() == "firefox":
            return (
                "Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:85.0) "
                "Gecko/20100101 Firefox/85.0"
            )
        else:
            return (
                "Mozilla/5.0 (X11; Linux x86_64) "
                "AppleWebKit/537.36 (KHTML, like Gecko) Chrome/88.0.4324.96 Safari/537.36"
            )

    def ev_inject(self, response):
        self._ev_dispatcher(response)



import base64
import pprint
import threading
import time
import zlib

from .constant import (
    ACTIVITY_STATE_KEY,
    AIR_QUALITY_KEY,
    AUDIO_ANALYTICS_KEY,
    AUDIO_DETECTED_KEY,
    AUDIO_POSITION_KEY,
    AUDIO_TRACK_KEY,
    BATTERY_KEY,
    BRIGHTNESS_KEY,
    CAPTURED_TODAY_KEY,
    CONNECTION_KEY,
    CRY_DETECTION_KEY,
    FLIP_KEY,
    FLOODLIGHT_BRIGHTNESS1_KEY,
    FLOODLIGHT_BRIGHTNESS2_KEY,
    FLOODLIGHT_KEY,
    HUMIDITY_KEY,
    IDLE_SNAPSHOT_PATH,
    LAMP_STATE_KEY,
    LAST_CAPTURE_KEY,
    LAST_IMAGE_DATA_KEY,
    LAST_IMAGE_KEY,
    LAST_IMAGE_SRC_KEY,
    LIGHT_BRIGHTNESS_KEY,
    LIGHT_MODE_KEY,
    MEDIA_COUNT_KEY,
    MEDIA_PLAYER_KEY,
    MEDIA_PLAYER_RESOURCE_ID,
    MEDIA_UPLOAD_KEY,
    MEDIA_UPLOAD_KEYS,
    MIRROR_KEY,
    MODEL_BABY,
    MODEL_ESSENTIAL,
    MODEL_ESSENTIAL_INDOOR,
    MODEL_PRO_2,
    MODEL_PRO_3,
    MODEL_PRO_3_FLOODLIGHT,
    MODEL_PRO_4,
    MODEL_ULTRA,
    MODEL_WIRED_VIDEO_DOORBELL,
    MODEL_WIREFREE_VIDEO_DOORBELL,
    MODEL_GO,
    MOTION_DETECTED_KEY,
    MOTION_SENS_KEY,
    NIGHTLIGHT_KEY,
    POWER_SAVE_KEY,
    PRIVACY_KEY,
    RECENT_ACTIVITY_KEY,
    RECENT_ACTIVITY_KEYS,
    RECORD_START_PATH,
    RECORD_STOP_PATH,
    RECORDING_STOPPED_KEY,
    SIGNAL_STR_KEY,
    SIREN_STATE_KEY,
    SNAPSHOT_KEY,
    SPOTLIGHT_BRIGHTNESS_KEY,
    SPOTLIGHT_KEY,
    STREAM_SNAPSHOT_KEY,
    STREAM_SNAPSHOT_PATH,
    STREAM_START_PATH,
    TEMPERATURE_KEY,
)
from .device import ArloChildDevice
from .util import http_get, http_get_img, the_epoch


class ArloCamera(ArloChildDevice):
    def __init__(self, name, arlo, attrs):
        super().__init__(name, arlo, attrs)
        self._recent = False
        self._recent_job = None
        self._cache_count = None
        self._cached_videos = None
        self._min_days_vdo_cache = self._arlo.cfg.library_days
        self._lock = threading.Condition()
        self._event = threading.Event()
        self._snapshot_time = the_epoch()
        self._stream_url = None
        # what user has requested locally
        self._user_requests = set()
        # what is keeping the stream open for us
        self._local_users = set()
        # what is triggered from elsewhere
        self._remote_users = set()

    def _parse_statistic(self, data, scale):
        """Parse binary statistics returned from the history API"""
        i = 0
        for byte in bytearray(data):
            i = (i << 8) + byte

        if i == 32768:
            return None

        if scale == 0:
            return i

        return float(i) / (scale * 10)

    def _decode_sensor_data(self, properties):
        """Decode, decompress, and parse the data from the history API"""
        b64_input = ""
        for s in properties.get("payload", []):
            # pylint: disable=consider-using-join
            b64_input += s
        if b64_input == "":
            return None

        decoded = base64.b64decode(b64_input)
        data = zlib.decompress(decoded)
        points = []
        i = 0

        while i < len(data):
            points.append(
                {
                    "timestamp": int(1e3 * self._parse_statistic(data[i : (i + 4)], 0)),
                    "temperature": self._parse_statistic(data[(i + 8) : (i + 10)], 1),
                    "humidity": self._parse_statistic(data[(i + 14) : (i + 16)], 1),
                    "airQuality": self._parse_statistic(data[(i + 20) : (i + 22)], 1),
                }
            )
            i += 22

        return points[-1]

    def _dump_activities(self, msg):
        self._arlo.debug(
            "{}::reqs='{}',local='{}',remote='{}'".format(
                msg,
                pprint.pformat(self._user_requests),
                pprint.pformat(self._local_users),
                pprint.pformat(self._remote_users),
            )
        )

    # Media library has updated, reload today's events.
    def _update_media(self):
        self._arlo.debug("reloading cache for " + self._name)
        count, videos = self._arlo.ml.videos_for(self)
        if videos:
            captured_today = len([video for video in videos if video.created_today])
            last_captured = videos[0].created_at_pretty(self._arlo.cfg.last_format)
            last_image = videos[0].thumbnail_url
        else:
            captured_today = 0
            last_captured = None
            last_image = None

        # update local copies
        with self._lock:
            self._cache_count = count
            self._cached_videos = videos

        # signal video up!
        self._save_and_do_callbacks(CAPTURED_TODAY_KEY, captured_today)
        if last_captured is not None:
            self._save_and_do_callbacks(LAST_CAPTURE_KEY, last_captured)
        self._do_callbacks(MEDIA_UPLOAD_KEY, True)

        # new snapshot?
        snapshot = self._arlo.ml.snapshot_for(self)
        if snapshot is not None:
            self._arlo.debug("snapshot updated for media " + self.name)
            self._save(SNAPSHOT_KEY, snapshot.image_url)
            self._arlo.bg.run_low(self._update_snapshot)

        # new image?
        if last_image is not None:
            self._arlo.debug("image updated for media " + self.name)
            self._save(LAST_IMAGE_KEY, last_image)
            self._arlo.bg.run_low(self._update_image)

    # Update last captured image.
    def _update_image(self):
        # Get image and date, if fails ignore
        img, date = http_get_img(self._load(LAST_IMAGE_KEY, None))
        if img is None:
            self._arlo.debug("failed to load image for " + self.name)
            return

        # Always make this the latest thumbnail image.
        if self._snapshot_time < date:
            self._arlo.debug("updating image for " + self.name)
            self._snapshot_time = date
            date = date.strftime(self._arlo.cfg.last_format)
            self._save_and_do_callbacks(LAST_IMAGE_SRC_KEY, "capture/" + date)
            self._save_and_do_callbacks(LAST_CAPTURE_KEY, date)
            self._save_and_do_callbacks(LAST_IMAGE_DATA_KEY, img)
        else:
            self._arlo.vdebug("ignoring image for " + self.name)

    # Update the last snapshot
    def _update_snapshot(self):
        # Get image and date, if fails ignore.
        img, date = http_get_img(self._load(SNAPSHOT_KEY, None))
        if img is None:
            self._arlo.debug("failed to load snapshot for " + self.name)
            return

        # Always make this the latest snapshot image.
        if self._snapshot_time < date:
            self._arlo.debug("updating snapshot for " + self.name)
            self._snapshot_time = date
            date = date.strftime(self._arlo.cfg.last_format)
            self._save_and_do_callbacks(LAST_IMAGE_SRC_KEY, "snapshot/" + date)
            self._save_and_do_callbacks(LAST_IMAGE_DATA_KEY, img)
            self._stop_snapshot()
        else:
            self._arlo.vdebug("ignoring snapshot for " + self.name)

    def _set_recent(self, timeo):
        with self._lock:
            self._recent = True
            self._arlo.bg.cancel(self._recent_job)
            self._recent_job = self._arlo.bg.run_in(self._clear_recent, timeo)
        self._arlo.debug("turning recent ON for " + self._name)
        self._do_callbacks(RECENT_ACTIVITY_KEY, True)

    def _clear_recent(self):
        with self._lock:
            self._recent = False
            self._recent_job = None
        self._arlo.debug("turning recent OFF for " + self._name)
        self._do_callbacks(RECENT_ACTIVITY_KEY, False)

    def _stop_snapshot(self):
        # Signal to anybody waiting.
        with self._lock:
            if not self.has_user_request("snapshot"):
                return
            self._user_requests.discard("snapshot")
            self._dump_activities("_stop_snapshot")
            self._lock.notify_all()

        # Stop based on how we were started.
        if not self.is_taking_idle_snapshot:
            self._stop_stream(stopping_for="snapshot")

        # Signal stop.
        self._arlo.debug("snapshot finished, re-signal real state")
        self._save_and_do_callbacks(
            ACTIVITY_STATE_KEY, self._load(ACTIVITY_STATE_KEY, "unknown")
        )

    def _queue_media_updates(self):
        for retry in self._arlo.cfg.media_retry:
            self._arlo.debug("queueing update in {}".format(retry))
            self._arlo.bg.run_in(
                self._arlo.ml.queue_update, retry, cb=self._update_media
            )

    def _mark_as_idle(self):
        """Camera has moved to idle.
        Either we did this or backend did this.
        """
        if self.has_any_local_users:
            self._arlo.debug("got a stream/recording stop")
            self._queue_media_updates()
            self._set_recent(self._arlo.cfg.recent_time)

        # Remove streaming from state
        self._arlo.debug("removing streaming activity state")
        with self._lock:
            self._local_users = set()
            self._remote_users = set()
            self._dump_activities("_event::idle")
            self._lock.notify_all()

    def _stop_activity(self):
        """Request the camera stop whatever it is doing and return to the idle state."""
        response = self._arlo.be.notify(
            base=self.base_station,
            body={
                "action": "set",
                "properties": {"activityState": "idle"},
                "publishResponse": True,
                "resource": self.resource_id,
            },
            wait_for="response",
        )
        if response is not None:
            self._mark_as_idle()

    def _start_stream(self, starting_for, user_agent=None):
        with self._lock:
            # Already streaming. Update sub-activity as needed.
            if self.has_any_local_users:
                self._local_users.add(starting_for)
                self._dump_activities("_start_stream")
                return self._stream_url

            # We can't start a stream if we are doing a straight snapshot.
            if self.is_taking_idle_snapshot:
                return None
            self._local_users.add(starting_for)
            self._dump_activities("_start_stream2")

        body = {
            "action": "set",
            "from": self.web_id,
            "properties": {
                "activityState": "startUserStream",
                "cameraId": self.device_id,
            },
            "publishResponse": True,
            "responseUrl": "",
            "resource": self.resource_id,
            "to": self.parent_id,
            "transId": self._arlo.be.gen_trans_id(),
        }

        headers = {"xcloudId": self.xcloud_id}
        if user_agent is not None:
            headers["User-Agent"] = self._arlo.be.user_agent(user_agent)

        self._stream_url = self._arlo.be.post(STREAM_START_PATH, body, headers=headers)
        if self._stream_url is not None:
            self._stream_url = self._stream_url["url"].replace("rtsp://", "rtsps://")
            self._arlo.debug("url={}".format(self._stream_url))
        else:
            with self._lock:
                self._local_users = set()
        return self._stream_url

    def _stop_stream(self, stopping_for="streaming"):
        with self._lock:
            self._local_users.discard(stopping_for)
            self._dump_activities("_stop_stream")
            if self.has_any_local_users:
                return
        self._stop_activity()

    def _event_handler(self, resource, event):
        self._arlo.debug(self.name + " CAMERA got one " + resource)

        # Stream has stopped or recording has stopped so new media is available.
        if resource == MEDIA_UPLOAD_KEY:

            # Look for easy keys.
            for key in MEDIA_UPLOAD_KEYS:
                value = event.get(key, None)
                if value is not None:
                    self._save_and_do_callbacks(key, value)

            # The last image thumbnail has changed. Queue an image or snapshot
            # update to download the image and process it.
            if LAST_IMAGE_KEY in event:
                if not self.is_taking_snapshot:
                    self._arlo.debug("{} -> thumbnail changed".format(self.name))
                    self._arlo.bg.run_low(self._update_image)
                else:
                    self._arlo.debug(
                        "{} -> snapshot(thumbnail) ready".format(self.name)
                    )
                    self._save(SNAPSHOT_KEY, event.get(LAST_IMAGE_KEY, ""))
                    self._arlo.bg.run_low(self._update_snapshot)

            # Recording has stopped so a new video is available. Queue an
            # media update, this could later trigger a snapshot or image
            # update.
            if event.get(RECORDING_STOPPED_KEY, False):
                self._arlo.debug("{} -> recording stopped".format(self.name))
                self._arlo.ml.queue_update(self._update_media)

            # Examine the URL passed; snapshots contain `/snapshots/` and
            # recordings contain `recordings`. For snapshot, save URL and queue
            # up an event to download and process it. We do nothing with the
            # recording for now, it will come in via a media update.
            value = event.get(STREAM_SNAPSHOT_KEY, "")
            if "/snapshots/" in value:
                self._arlo.debug("{} -> snapshot1 ready".format(self.name))
                self._save(SNAPSHOT_KEY, value)
                self._arlo.bg.run_low(self._update_snapshot)
            if "/recordings/" in value:
                self._arlo.debug("{} -> new recording ready".format(self.name))

            # Something just happened.
            self._set_recent(self._arlo.cfg.recent_time)

            return

        # Camera Activity State
        activity = event.get("properties", {}).get("activityState", "unknown")

        # Camera has gone idle.
        if activity == "idle":
            self._mark_as_idle()

        # Camera is active. If we don't know about it then update our status.
        if activity == "fullFrameSnapshot":
            with self._lock:
                if not self.has_user_request("snapshot"):
                    self._remote_users.add("snapshot")
                    #  if not self.has_any_local_users:
                    #  self._local_users.add("remote")
                self._dump_activities("_event::snap")
        if activity == "alertStreamActive":
            with self._lock:
                if not self.has_user_request("recording"):
                    self._remote_users.add("recording")
                    if not self.has_any_local_users:
                        self._local_users.add("remote")
                self._lock.notify_all()
                self._dump_activities("_event::record")
        if activity == "userStreamActive":
            with self._lock:
                if not self.has_user_request("streaming"):
                    self._remote_users.add("streaming")
                    if not self.has_any_local_users:
                        self._local_users.add("remote")
                self._lock.notify_all()
                self._dump_activities("_event::stream")

        # Snapshot is updated. Queue retrieval.
        if event.get("action", "") == "fullFrameSnapshotAvailable":
            value = event.get("properties", {}).get(
                "presignedFullFrameSnapshotUrl", None
            )
            if value is not None:
                self._arlo.debug("{} -> snapshot2 ready".format(self.name))
                self._save(SNAPSHOT_KEY, value)
                self._arlo.bg.run_low(self._update_snapshot)

        # Non subscription...
        if event.get("action", "") == "lastImageSnapshotAvailable":
            value = event.get("properties", {}).get("presignedLastImageUrl", None)
            if value is not None:
                self._arlo.debug("{} -> snapshot3 ready".format(self.name))
                self._save(SNAPSHOT_KEY, value)
                self._arlo.bg.run_low(self._update_snapshot)

        # Ambient sensors update, decode and push changes.
        if resource.endswith("/ambientSensors/history"):
            data = self._decode_sensor_data(event.get("properties", {}))
            if data is not None:
                self._save_and_do_callbacks("temperature", data.get("temperature"))
                self._save_and_do_callbacks("humidity", data.get("humidity"))
                self._save_and_do_callbacks("airQuality", data.get("airQuality"))

        # Properties settings.
        properties = event.get("properties", {})

        # Anything to trip recent activity?
        for key in properties:
            if key in RECENT_ACTIVITY_KEYS:
                self._arlo.debug("recent activity key")
                self._set_recent(self._arlo.cfg.recent_time)

        # Local record stopped, try and trip and update.
        if properties.get("localRecordingActive", True) is False:
            self._arlo.debug("local recording stopped, updating media")
            self._queue_media_updates()

        # Night light status.
        nightlight = properties.get(NIGHTLIGHT_KEY, None)
        if nightlight is not None:
            self._arlo.debug(
                "got a night light {}".format(nightlight.get("enabled", False))
            )
            if nightlight.get("enabled", False) is True:
                self._save_and_do_callbacks(LAMP_STATE_KEY, "on")
            else:
                self._save_and_do_callbacks(LAMP_STATE_KEY, "off")

            brightness = nightlight.get("brightness")
            if brightness is not None:
                self._save_and_do_callbacks(LIGHT_BRIGHTNESS_KEY, brightness)

            mode = nightlight.get("mode")
            if mode is not None:
                rgb = nightlight.get("rgb")
                temperature = nightlight.get("temperature")

                light_mode = {"mode": mode}

                if rgb is not None:
                    light_mode["rgb"] = rgb
                if temperature is not None:
                    light_mode["temperature"] = temperature

                self._save_and_do_callbacks(LIGHT_MODE_KEY, light_mode)

        # Spotlight status.
        spotlight = properties.get(SPOTLIGHT_KEY, None)
        if spotlight is not None:
            self._arlo.debug(
                "got a spotlight {}".format(spotlight.get("enabled", False))
            )
            if spotlight.get("enabled", False) is True:
                self._save_and_do_callbacks(SPOTLIGHT_KEY, "on")
            else:
                self._save_and_do_callbacks(SPOTLIGHT_KEY, "off")

            brightness = spotlight.get("intensity")
            if brightness is not None:
                self._save_and_do_callbacks(SPOTLIGHT_BRIGHTNESS_KEY, brightness)

        # Floodlight status.
        floodlight = properties.get(FLOODLIGHT_KEY, None)
        if floodlight is not None:
            self._arlo.debug("got a flood light {}".format(floodlight.get("on", False)))
            self._save_and_do_callbacks(FLOODLIGHT_KEY, floodlight)

        # Audio analytics.
        audioanalytics = properties.get(AUDIO_ANALYTICS_KEY, None)
        if audioanalytics is not None:
            triggered = audioanalytics.get(CRY_DETECTION_KEY, {}).get(
                "triggered", False
            )
            self._save_and_do_callbacks(CRY_DETECTION_KEY, triggered)

        # Pass event to lower level.
        super()._event_handler(resource, event)

    @property
    def resource_type(self):
        return "cameras"

    @property
    def last_thumbnail(self):
        """Returns the URL of the last image as reported by Arlo."""
        return self._load(LAST_IMAGE_KEY, None)

    @property
    def last_snapshot(self):
        """Returns the URL of the last snapshot as reported by Arlo."""
        return self._load(SNAPSHOT_KEY, None)

    @property
    def last_image(self):
        """Returns the URL of the last snapshot or image taken.

        Will pick snapshot or image based on most recently updated.
        """
        image = None
        if self.last_image_source.startswith("snapshot/"):
            image = self.last_snapshot
        if image is None:
            image = self.last_thumbnail
        return image

    @property
    def last_image_from_cache(self):
        """Returns the last image or snapshot in binary format.

        :return: Binary reprsensation of the last image.
        :rtype: bytearray
        """
        return self._load(LAST_IMAGE_DATA_KEY, self._arlo.blank_image)

    @property
    def last_image_source(self):
        """Returns a string describing what triggered the last image capture.

        Currently either `capture/${date}` or `snapshot/${date}`.
        """
        return self._load(LAST_IMAGE_SRC_KEY, "")

    @property
    def last_video(self):
        """Returns a video object describing the last captured video.

        :return: Video object or `None` if no videos present.
        :rtype: ArloVideo
        """
        with self._lock:
            if self._cached_videos:
                return self._cached_videos[0]
        return None

    def last_n_videos(self, count):
        """Returns the last count video objects describing the last captured videos.

        :return: `count` video objects or `None` if no videos present.
        :rtype: list(ArloVideo)
        """
        with self._lock:
            if self._cached_videos:
                return self._cached_videos[:count]
        return []

    @property
    def last_capture(self):
        """Returns a date string showing when the last video was captured.

        It uses the format returned by `last_capture_date_format`.
        """
        return self._load(LAST_CAPTURE_KEY, None)

    @property
    def last_capture_date_format(self):
        """Returns a date format string used by the last_capture function.

        You can set this value in the parameters passed to PyArlo.
        """
        return self._arlo.cfg.last_format

    @property
    def brightness(self):
        """Returns the camera brightness setting."""
        return self._load(BRIGHTNESS_KEY, None)

    @brightness.setter
    def brightness(self, brightness):
        """
        NOTE: Brightness is between -2 and 2 in increments of 1 (-2, -1, 0, 1, 2).
        Setting it to an invalid value has no effect.
        """
        body = {
            "action": "set",
            "resource": self.resource_id,
            "publishResponse": True,
            "properties": {"brightness": brightness},
        }
        self._arlo.be.notify(base=self, body=body)

    @property
    def flip_state(self):
        """Returns `True` if the camera is flipped, `False` otherwise."""
        return self._load(FLIP_KEY, None)

    @property
    def mirror_state(self):
        """Returns `True` if the camera is mirrored, `False` otherwise."""
        return self._load(MIRROR_KEY, None)

    @property
    def motion_detection_sensitivity(self):
        """Returns the camera motion sensitivity setting."""
        return self._load(MOTION_SENS_KEY, None)

    @property
    def powersave_mode(self):
        """Returns `True` if the camera is on power save mode, `False` otherwise."""
        return self._load(POWER_SAVE_KEY, None)

    @property
    def unseen_videos(self):
        """Returns the camera unseen video count."""
        return self._load(MEDIA_COUNT_KEY, 0)

    @property
    def captured_today(self):
        """Returns the number of videos captured today."""
        return self._load(CAPTURED_TODAY_KEY, 0)

    @property
    def min_days_vdo_cache(self):
        return self._min_days_vdo_cache

    @min_days_vdo_cache.setter
    def min_days_vdo_cache(self, value):
        self._min_days_vdo_cache = value

    def update_media(self, wait=None):
        """Requests latest list of recordings from the backend server.

        :param wait if True then wait for completion, if False then don't wait,
        if None then use synchronous_mode setting.

        Reloads the videos library from Arlo.
        """
        if wait is None:
            wait = self._arlo.cfg.synchronous_mode
        if wait:
            self._arlo.debug("doing media update")
            self._update_media()
        else:
            self._arlo.debug("queueing media update")
            self._arlo.bg.run_low(self._update_media)

    def update_last_image(self, wait=None):
        """Requests last thumbnail from the backend server.

        :param wait if True then wait for completion, if False then don't wait,
        if None then use synchronous_mode setting.

        Updates the last image.
        """
        if wait is None:
            wait = self._arlo.cfg.synchronous_mode
        if wait:
            self._arlo.debug("doing image update")
            self._update_image()
        else:
            self._arlo.debug("queueing image update")
            self._arlo.bg.run_low(self._update_image)

    def update_ambient_sensors(self):
        """Requests the latest temperature, humidity and air quality settings.

        Queues a job that requests the info from Arlo.
        """
        if self.model_id == MODEL_BABY:
            self._arlo.be.notify(
                base=self.base_station,
                body={
                    "action": "get",
                    "resource": "cameras/{}/ambientSensors/history".format(
                        self.device_id
                    ),
                    "publishResponse": False,
                },
            )

    def _take_streaming_snapshot(self):
        body = {
            "xcloudId": self.xcloud_id,
            "parentId": self.parent_id,
            "deviceId": self.device_id,
            "olsonTimeZone": self.timezone,
        }
        self._arlo.bg.run(
            self._arlo.be.post,
            path=STREAM_SNAPSHOT_PATH,
            params=body,
            headers={"xcloudId": self.xcloud_id},
        )

    def _take_idle_snapshot(self):
        body = {
            "action": "set",
            "from": self.web_id,
            "properties": {"activityState": "fullFrameSnapshot"},
            "publishResponse": True,
            "resource": self.resource_id,
            "to": self.parent_id,
            "transId": self._arlo.be.gen_trans_id(),
        }
        self._arlo.bg.run(
            self._arlo.be.post,
            path=IDLE_SNAPSHOT_PATH,
            params=body,
            headers={"xcloudId": self.xcloud_id},
        )

    def request_snapshot(self):
        """Requests a snapshot from the camera without blocking.

        The snapshot can be handled with callbacks registered to
        LAST_IMAGE_SRC_KEY - lastImageSource starting with snapshot/, or capture/
        LAST_IMAGE_DATA_KEY - presignedLastImageData containing the image data.
        """
        with self._lock:
            if self.has_user_request("snapshot"):
                return
            stream_snapshot = self.has_any_local_users
            self._user_requests.add("snapshot")
            self._dump_activities("request_snapshot")
            snapshot_running = self.has_remote_user("snapshot")

        self._save_and_do_callbacks(ACTIVITY_STATE_KEY, "fullFrameSnapshot")
        if not snapshot_running:
            if stream_snapshot:
                self._arlo.debug("streaming/recording snapshot")
                self._take_streaming_snapshot()
                if self._arlo.cfg.stream_snapshot_stop > 0:
                    self._arlo.debug(
                        "queing stream stop in {}".format(
                            self._arlo.cfg.stream_snapshot_stop
                        )
                    )
                    self._arlo.bg.run_in(
                        self._stop_stream,
                        self._arlo.cfg.stream_snapshot_stop,
                        stopping_for="snapshot",
                    )
            else:
                self._arlo.debug("idle snapshot")
                self._take_idle_snapshot()

        for check in self._arlo.cfg.snapshot_checks:
            self._arlo.debug("queueing snapshot check in {}".format(check))
            self._arlo.bg.run_in(
                self._arlo.ml.queue_update, check, cb=self._update_media
            )

        self._arlo.vdebug("handle dodgy cameras")
        self._arlo.bg.run_in(self._stop_snapshot, self._arlo.cfg.snapshot_timeout)

    def get_snapshot(self, timeout=60):
        """Gets a snapshot from the camera and returns it.

        :param timeout: how long to wait, in seconds, before stopping the snapshot attempt
        :return: a binary represention of the image, or the last image if snapshot timed out
        :rtype: bytearray
        """
        self.request_snapshot()

        mnow = time.monotonic()
        mend = mnow + timeout
        with self._lock:
            while mnow < mend and self.has_user_request("snapshot"):
                self._lock.wait(mend - mnow)
                mnow = time.monotonic()
        self._arlo.debug("finished snapshot")
        return self.last_image_from_cache

    @property
    def is_taking_snapshot(self):
        """Returns `True` if camera is taking a snapshot, `False` otherwise.

        Snapshot can be started from anywhere.
        """
        return self.has_user_request("snapshot") or self.has_remote_user("snapshot")

    @property
    def is_taking_idle_snapshot(self):
        """Returns `True` if camera is taking a non-streaming snapshot, `False`
        otherwise.
        """
        return self.is_taking_snapshot and not self.has_any_local_users

    @property
    def is_recording(self):
        """Returns `True` if camera is recording a video, `False` otherwise.

        Recording can be started from anywhere.
        """
        return self.has_user_request("recording") or self.has_remote_user("recording")

    @property
    def is_streaming(self):
        """Returns `True` if camera is streaming a video, `False` otherwise.

        Stream has to be started locally.
        """
        return self.has_user_request("streaming") or self.has_remote_user("streaming")

    def has_user_request(self, activity):
        return activity in self._user_requests

    @property
    def has_any_user_requests(self):
        return len(self._user_requests) != 0

    def has_local_user(self, activity):
        return activity in self._local_users

    @property
    def has_any_local_users(self):
        return len(self._local_users) != 0

    def has_remote_user(self, activity):
        return activity in self._remote_users

    @property
    def has_any_remote_users(self):
        return len(self._remote_users) != 0

    def has_activity(self, activity):
        """Returns `True` is camera is performing a particular activity,
        `False` otherwise.
        """
        return (
            self.has_user_request(activity)
            or self.has_local_user(activity)
            or self.has_remote_user(activity)
        )

    @property
    def was_recently_active(self):
        """Returns `True` if camera was recently active, `False` otherwise."""
        return self._recent

    @property
    def state(self):
        """Returns the camera's current state."""
        if not self.is_on:
            return "off"
        if self.has_local_user("snapshot"):
            if self.has_local_user("recording"):
                return "recording + snapshot"
            if self.has_local_user("streaming"):
                return "streaming + snapshot"
            return "taking snapshot"
        if self.has_activity("recording"):
            return "recording"
        if self.has_activity("streaming"):
            return "streaming"
        if self.was_recently_active:
            return "recently active"
        return super().state

    def get_stream(self, user_agent=None):
        """Start a stream and return the URL for it.

        Code does nothing with the url, it's up to you to pass the url to something.

        The stream will stop if nothing connects to it within 30 seconds.
        """
        return self._start_stream("streaming", user_agent)

    def start_stream(self, user_agent=None):
        """Start a stream and return the URL for it.

        Code does nothing with the url, it's up to you to pass the url to something.

        The stream will stop if nothing connects to it within 30 seconds.
        """
        return self._start_stream("streaming", user_agent)

    def start_snapshot_stream(self, user_agent=None):
        return self._start_stream("snapshot", user_agent)

    def start_recording_stream(self, user_agent=None):
        return self._start_stream("recording", user_agent)

    def stop_stream(self):
        self._stop_stream("streaming")

    def stop_snapshot_stream(self):
        self._stop_stream("snapshot")

    def stop_recording_stream(self):
        self._stop_stream("recording")

    def wait_for_user_stream(self, timeout=15):
        self._arlo.debug("waiting for stream")
        mnow = time.monotonic()
        mend = mnow + timeout
        with self._lock:
            while mnow < mend and not self.has_remote_user("streaming"):
                self._lock.wait(mend - mnow)
                mnow = time.monotonic()
            active = self.has_remote_user("streaming")

        # Is active, give a small delay to get going.
        if active:
            self._arlo.debug("delaying stream start")
            self._event.wait(self._arlo.cfg.user_stream_delay)
        return active

    def get_video(self):
        """Download and return the last recorded video.

        **Note:** Prefer getting the url and downloading it yourself.
        """
        video = self.last_video
        if video is not None:
            return http_get(video.video_url)
        return None

    def stop_activity(self):
        """Request the camera stop whatever it is doing and return to the idle state."""
        # has_any_activity
        self._stop_activity()
        return True

    def start_recording(self, duration=None):
        """Request the camera start recording.

        :param duration: seconds for recording to run, `None` means no stopping.

        **Note:** Arlo will stop the recording after 30 seconds if nothing
        connects to the stream.
        **Note:** Arlo will stop the recording after 30 minutes anyway.
        """
        with self._lock:
            if not self.has_any_local_users:
                return None
            if self.has_user_request("recording"):
                return self._stream_url
            self._user_requests.add("recording")
            self._dump_activities("start_recording")

        body = {
            "parentId": self.parent_id,
            "deviceId": self.device_id,
            "olsonTimeZone": self.timezone,
        }
        self._arlo.debug("starting recording")
        self._save_and_do_callbacks(ACTIVITY_STATE_KEY, "alertStreamActive")
        self._arlo.bg.run(
            self._arlo.be.post,
            path=RECORD_START_PATH,
            params=body,
            headers={"xcloudId": self.xcloud_id},
        )

        # Queue up stop.
        if duration is not None:
            self._arlo.debug("queueing stop")
            self._arlo.bg.run_in(self.stop_recording, duration)

        return self._stream_url

    def stop_recording(self):
        """Request the camera stop recording."""
        with self._lock:
            if not self.has_user_request("recording") and not self.has_remote_user(
                "recording"
            ):
                return
            self._user_requests.discard("recording")
            self._dump_activities("stop_recording")

        body = {
            "parentId": self.parent_id,
            "deviceId": self.device_id,
        }
        self._arlo.debug("stopping recording")
        self._arlo.bg.run(
            self._arlo.be.post,
            path=RECORD_STOP_PATH,
            params=body,
            headers={"xcloudId": self.xcloud_id},
        )

        # stop stream
        self._arlo.bg.run_in(self.stop_recording_stream, 1)

    @property
    def _siren_resource_id(self):
        return "siren/{}".format(self.device_id)

    @property
    def siren_state(self):
        return self._load(SIREN_STATE_KEY, "off")

    def siren_on(self, duration=300, volume=8):
        """Turn camera siren on.

        Does nothing if camera doesn't support sirens.

        :param duration: how long, in seconds, to sound for
        :param volume: how long, from 1 to 8, to sound
        """
        body = {
            "action": "set",
            "resource": self._siren_resource_id,
            "publishResponse": True,
            "properties": {
                "sirenState": "on",
                "duration": int(duration),
                "volume": int(volume),
                "pattern": "alarm",
            },
        }
        self._arlo.be.notify(base=self, body=body)

    def siren_off(self):
        """Turn camera siren off.

        Does nothing if camera doesn't support sirens.
        """
        body = {
            "action": "set",
            "resource": self._siren_resource_id,
            "publishResponse": True,
            "properties": {"sirenState": "off"},
        }
        self._arlo.be.notify(base=self, body=body)

    @property
    def is_on(self):
        """Returns `True` if the camera turned on."""
        return not self._load(PRIVACY_KEY, False)

    def turn_on(self):
        """Turn the camera on."""
        body = {
            "action": "set",
            "resource": self.resource_id,
            "publishResponse": True,
            "properties": {"privacyActive": False},
        }
        self._arlo.be.notify(base=self.base_station, body=body)

    def turn_off(self):
        """Turn the camera off."""
        body = {
            "action": "set",
            "resource": self.resource_id,
            "publishResponse": True,
            "properties": {"privacyActive": True},
        }
        self._arlo.be.notify(base=self.base_station, body=body)

    def get_audio_playback_status(self):
        """Gets the current playback status and available track list"""
        body = {"action": "get", "publishResponse": True, "resource": "audioPlayback"}
        self._arlo.be.notify(base=self, body=body)

    def play_track(self, track_id=None, position=0):
        """Play the track. A track ID of None will resume playing the current
        track.

        :param track_id: track id
        :param position: position in the track
        """
        body = {
            "publishResponse": True,
            "resource": MEDIA_PLAYER_RESOURCE_ID,
        }

        if track_id is not None:
            body.update(
                {
                    "action": "playTrack",
                    "properties": {
                        AUDIO_TRACK_KEY: track_id,
                        AUDIO_POSITION_KEY: position,
                    },
                }
            )
        else:
            body.update(
                {
                    "action": "play",
                }
            )
        self._arlo.be.notify(base=self, body=body)

    def pause_track(self):
        """Pause the playing track."""
        body = {
            "action": "pause",
            "publishResponse": True,
            "resource": MEDIA_PLAYER_RESOURCE_ID,
        }
        self._arlo.be.notify(base=self, body=body)

    def previous_track(self):
        """Skips to the previous track in the playlist."""
        body = {
            "action": "prevTrack",
            "publishResponse": True,
            "resource": MEDIA_PLAYER_RESOURCE_ID,
        }
        self._arlo.be.notify(base=self, body=body)

    def next_track(self):
        """Skips to the next track in the playlist."""
        body = {
            "action": "nextTrack",
            "publishResponse": True,
            "resource": MEDIA_PLAYER_RESOURCE_ID,
        }
        self._arlo.be.notify(base=self, body=body)

    def set_music_loop_mode_continuous(self):
        """Sets the music loop mode to repeat the entire playlist."""
        body = {
            "action": "set",
            "publishResponse": True,
            "resource": "audioPlayback/config",
            "properties": {"config": {"loopbackMode": "continuous"}},
        }
        self._arlo.be.notify(base=self, body=body)

    def set_music_loop_mode_single(self):
        """Sets the music loop mode to repeat the current track."""
        body = {
            "action": "set",
            "publishResponse": True,
            "resource": "audioPlayback/config",
            "properties": {"config": {"loopbackMode": "singleTrack"}},
        }
        self._arlo.be.notify(base=self, body=body)

    def set_shuffle(self, shuffle=True):
        """Sets playback to shuffle.

        :param shuffle: `True` to turn on shuffle.
        """
        body = {
            "action": "set",
            "publishResponse": True,
            "resource": "audioPlayback/config",
            "properties": {"config": {"shuffleActive": shuffle}},
        }
        self._arlo.be.notify(base=self, body=body)

    def set_volume(self, mute=False, volume=50):
        """Sets the music volume.

        :param mute: `True` to mute the volume.
        :param volume: set volume (0-100)
        """
        body = {
            "action": "set",
            "publishResponse": True,
            "resource": self.resource_id,
            "properties": {"speaker": {"mute": mute, "volume": volume}},
        }
        self._arlo.be.notify(base=self, body=body)

    def _set_nightlight_properties(self, properties):
        self._arlo.debug(
            "{}: setting nightlight properties: {}".format(self._name, properties)
        )
        self._arlo.be.notify(
            base=self.base_station,
            body={
                "action": "set",
                "properties": {"nightLight": properties},
                "publishResponse": True,
                "resource": self.resource_id,
            },
        )
        return True

    def nightlight_on(self):
        """Turns the nightlight on."""
        return self._set_nightlight_properties({"enabled": True})

    def nightlight_off(self):
        """Turns the nightlight off."""
        return self._set_nightlight_properties({"enabled": False})

    def set_nightlight_brightness(self, brightness):
        """Sets the nightlight brightness.

        :param brightness: brightness (0-255)
        """
        return self._set_nightlight_properties({"brightness": brightness})

    def set_nightlight_rgb(self, red=255, green=255, blue=255):
        """Turns the nightlight color to the specified RGB value.

        :param red: red value
        :param green: green value
        :param blue: blue value
        """
        return self._set_nightlight_properties(
            {"mode": "rgb", "rgb": {"red": red, "green": green, "blue": blue}}
        )

    def set_nightlight_color_temperature(self, temperature):
        """Turns the nightlight to the specified Kelvin color temperature.

        :param temperature: temperature, in Kelvin
        """
        return self._set_nightlight_properties(
            {"mode": "temperature", "temperature": str(temperature)}
        )

    def set_nightlight_mode(self, mode):
        """Turns the nightlight to a particular mode.

        :param mode: either `rgb`, `temperature` or `rainbow`
        :return:
        """
        return self._set_nightlight_properties({"mode": mode})

    def _set_spotlight_properties(self, properties):
        self._arlo.debug(
            "{}: setting spotlight properties: {}".format(self._name, properties)
        )
        self._arlo.be.notify(
            base=self.base_station,
            body={
                "action": "set",
                "properties": {"spotlight": properties},
                "publishResponse": True,
                "resource": self.resource_id,
            },
        )
        return True

    def set_spotlight_on(self):
        """Turns the spotlight on"""
        return self._set_spotlight_properties({"enabled": True})

    def set_spotlight_off(self):
        """Turns the spotlight off"""
        return self._set_spotlight_properties({"enabled": False})

    def set_spotlight_brightness(self, brightness):
        """Sets the nightlight brightness.

        :param brightness: brightness (0-255)
        """
        # Note: Intensity is 0-100 scale, which we map from 0-255 to
        #       provide an API consistent with nightlight brightness
        return self._set_spotlight_properties({"intensity": (brightness / 255 * 100)})

    def _set_floodlight_properties(self, properties):
        self._arlo.debug(
            "{}: setting floodlight properties: {}".format(self._name, properties)
        )
        self._arlo.be.notify(
            base=self.base_station,
            body={
                "action": "set",
                "properties": {"floodlight": properties},
                "publishResponse": True,
                "resource": self.resource_id,
            },
        )
        return True

    def floodlight_on(self):
        """Turns the floodlight on."""
        return self._set_floodlight_properties({"on": True})

    def floodlight_off(self):
        """Turns the floodlight off."""
        return self._set_floodlight_properties({"on": False})

    def set_floodlight_brightness(self, brightness):
        """Turns the floodlight brightness value (0-255)."""
        percentage = int(brightness / 255 * 100)
        return self._set_floodlight_properties(
            {
                FLOODLIGHT_BRIGHTNESS1_KEY: percentage,
                FLOODLIGHT_BRIGHTNESS2_KEY: percentage,
            }
        )

    def has_capability(self, cap):
        if cap in (BATTERY_KEY,):
            if self.model_id.startswith(MODEL_ESSENTIAL_INDOOR):
                return False
            else:
                return True
        if cap in (MOTION_DETECTED_KEY, SIGNAL_STR_KEY):
            return True
        if cap in (LAST_CAPTURE_KEY, CAPTURED_TODAY_KEY, RECENT_ACTIVITY_KEY):
            return True
        if cap in (AUDIO_DETECTED_KEY,):
            if self.model_id.startswith(
                (
                    MODEL_ESSENTIAL,
                    MODEL_ESSENTIAL_INDOOR,
                    MODEL_PRO_2,
                    MODEL_PRO_3,
                    MODEL_PRO_3_FLOODLIGHT,
                    MODEL_PRO_4,
                    MODEL_ULTRA,
                    MODEL_GO,
                    MODEL_BABY,
                )
            ):
                return True
            if self.device_type.startswith("arloq"):
                return True
        if cap in (SIREN_STATE_KEY,):
            if self.model_id.startswith(
                (
                    MODEL_ESSENTIAL,
                    MODEL_ESSENTIAL_INDOOR,
                    MODEL_PRO_3,
                    MODEL_PRO_3_FLOODLIGHT,
                    MODEL_PRO_4,
                    MODEL_ULTRA,
                    MODEL_WIREFREE_VIDEO_DOORBELL,
                )
            ):
                return True
        if cap in (SPOTLIGHT_KEY,):
            if self.model_id.startswith(
                (MODEL_ESSENTIAL, MODEL_PRO_3, MODEL_PRO_4, MODEL_ULTRA)
            ):
                return True
        if cap in (TEMPERATURE_KEY, HUMIDITY_KEY, AIR_QUALITY_KEY):
            if self.model_id.startswith(MODEL_BABY):
                return True
        if cap in (MEDIA_PLAYER_KEY, NIGHTLIGHT_KEY, CRY_DETECTION_KEY):
            if self.model_id.startswith(MODEL_BABY):
                return True
        if cap in (FLOODLIGHT_KEY,):
            if self.model_id.startswith(MODEL_PRO_3_FLOODLIGHT):
                return True
        if cap in (CONNECTION_KEY,):
            # These devices are their own base stations so don't re-add connection key.
            if self.model_id.startswith(
                (
                    MODEL_BABY,
                    MODEL_PRO_3_FLOODLIGHT,
                    MODEL_PRO_4,
                    MODEL_ESSENTIAL,
                    MODEL_WIRED_VIDEO_DOORBELL,
                    MODEL_WIREFREE_VIDEO_DOORBELL,
                    MODEL_ESSENTIAL_INDOOR,
                    MODEL_GO,
                )
            ):
                return False
            if self.device_type in ("arloq", "arloqs"):
                return False
        return super().has_capability(cap)



import base64
import time
from datetime import datetime, timezone

import requests


def utc_to_local(utc_dt):
    return utc_dt.replace(tzinfo=timezone.utc).astimezone(tz=None)


def the_epoch():
    return utc_to_local(datetime.fromtimestamp(0, tz=timezone.utc))


def arlotime_to_time(timestamp):
    """Convert Arlo timestamp to Unix timestamp."""
    return int(timestamp / 1000)


def arlotime_to_datetime(timestamp):
    """Convert Arlo timestamp to Python datetime."""
    return utc_to_local(datetime.fromtimestamp(int(timestamp / 1000), tz=timezone.utc))


def arlotime_strftime(timestamp, date_format="%Y-%m-%dT%H:%M:%S"):
    """Convert Arlo timestamp to time string."""
    return arlotime_to_datetime(timestamp).strftime(date_format)


def time_to_arlotime(timestamp=None):
    """Convert Unix timestamp to Arlo timestamp."""
    if timestamp is None:
        timestamp = time.time()
    return int(timestamp * 1000)


def now_strftime(date_format="%Y-%m-%dT%H:%M:%S"):
    """Convert now to time string."""
    return datetime.now().strftime(date_format)


def httptime_to_datetime(http_timestamp):
    """Convert HTTP timestamp to Python datetime."""
    return utc_to_local(datetime.strptime(http_timestamp, "%a, %d %b %Y %H:%M:%S GMT"))


def httptime_strftime(http_timestamp, date_format="%Y-%m-%dT%H:%M:%S"):
    """Convert HTTP timestamp to time string."""
    return httptime_to_datetime(http_timestamp).strftime(date_format)


def _http_get(url):
    """Download HTTP data."""

    if url is None:
        return None

    try:
        ret = requests.get(url)
    except requests.exceptions.SSLError:
        return None
    except Exception:
        return None

    if ret.status_code != 200:
        return None
    return ret


def http_get(url, filename=None):
    """Download HTTP data."""

    ret = _http_get(url)
    if ret is None:
        return False

    if filename is None:
        return ret.content

    with open(filename, "wb") as data:
        data.write(ret.content)
    return True


def http_get_img(url):
    """Download HTTP image data."""

    ret = _http_get(url)
    if ret is None:
        return None, datetime.now()

    date = ret.headers.get("Last-Modified", None)
    if date is not None:
        date = httptime_to_datetime(date)
    else:
        date = datetime.now()
    return ret.content, date


def http_stream(url, chunk=4096):
    """Generate stream for a given record video.

    :param url: url of stream to read
    :param chunk: chunk bytes to read per time
    :returns generator object
    """
    ret = requests.get(url, stream=True)
    ret.raise_for_status()
    for data in ret.iter_content(chunk):
        yield data


def rgb_to_hex(rgb):
    """Convert HA color to Arlo color."""
    return "#{:02x}{:02x}{:02x}".format(rgb[0], rgb[1], rgb[2])


def hex_to_rgb(h):
    """Convert Arlo color to HA color."""
    return {"red": int(h[1:3], 16), "green": int(h[3:5], 16), "blue": int(h[5:7], 16)}


def to_b64(in_str):
    """Convert a string into a base64 string."""
    return base64.b64encode(in_str.encode()).decode()



import threading

from unidecode import unidecode

from .constant import (
    BATTERY_KEY,
    BATTERY_TECH_KEY,
    CHARGER_KEY,
    CHARGING_KEY,
    CONNECTION_KEY,
    CONNECTIVITY_KEY,
    DEVICE_KEYS,
    RESOURCE_KEYS,
    RESOURCE_UPDATE_KEYS,
    SIGNAL_STR_KEY,
    XCLOUD_ID_KEY,
)


class ArloDevice(object):
    """Base class for all Arlo devices.

    Has code to handle providing common attributes and comment event handling.
    """

    def __init__(self, name, arlo, attrs):
        self._name = name
        self._arlo = arlo
        self._attrs = attrs

        self._lock = threading.Lock()
        self._attr_cbs_ = []

        # Basic IDs.
        self._device_id = attrs.get("deviceId", None)
        self._device_type = attrs.get("deviceType", "unknown")
        self._unique_id = attrs.get("uniqueId", None)

        # We save this here but only expose it directly in the ArloChild class.
        # Some devices are their own parents and we need to know that at the ArloDevice
        # or ArloChild class so we leave this here as a short cut.
        self._parent_id = attrs.get("parentId", None)

        # Activities. Used by camera for now but made available to all.
        self._activities = {}

        # Build initial values... attrs is device state
        for key in DEVICE_KEYS:
            value = attrs.get(key, None)
            if value is not None:
                self._save(key, value)

        # add a listener
        self._arlo.be.add_listener(self, self._event_handler)

    def __repr__(self):
        # Representation string of object.
        return "<{0}:{1}:{2}>".format(
            self.__class__.__name__, self._device_type, self._name
        )

    def _to_storage_key(self, attr):
        # Build a key incorporating the type!
        if isinstance(attr, list):
            return [self.__class__.__name__, self._device_id] + attr
        else:
            return [self.__class__.__name__, self._device_id, attr]

    def _event_handler(self, resource, event):
        self._arlo.vdebug("{}: got {} event **".format(self.name, resource))

        # Find properties. Event either contains a item called properties or it
        # is the whole thing.
        props = event.get("properties", event)

        # Save out new values.
        for key in props:
            if key in RESOURCE_KEYS or key in RESOURCE_UPDATE_KEYS:
                value = props.get(key, None)
                if value is not None:
                    self._save_and_do_callbacks(key, value)

    def _do_callbacks(self, attr, value):
        cbs = []
        with self._lock:
            for watch, cb in self._attr_cbs_:
                if watch == attr or watch == "*":
                    cbs.append(cb)
        for cb in cbs:
            cb(self, attr, value)

    def _save(self, attr, value):
        self._arlo.st.set(self._to_storage_key(attr), value)

    def _save_and_do_callbacks(self, attr, value):
        # TODO only care if it changes?
        if self._load(attr) != value:
            self._arlo.debug(f"{attr} now= {value}")
            self._save(attr, value)
            self._do_callbacks(attr, value)
        else:
            self._arlo.debug(f"{attr} still= {value}")

    def _load(self, attr, default=None):
        return self._arlo.st.get(self._to_storage_key(attr), default)

    def _load_matching(self, attr, default=None):
        return self._arlo.st.get_matching(self._to_storage_key(attr), default)

    @property
    def entity_id(self):
        if self._arlo.cfg.serial_ids:
            return self.device_id
        elif self._arlo.cfg.no_unicode_squash:
            return self.name.lower().replace(" ", "_")
        else:
            return unidecode(self.name.lower().replace(" ", "_"))

    @property
    def name(self):
        """Returns the device name."""
        return self._name

    @property
    def device_id(self):
        """Returns the device's id."""
        return self._device_id

    @property
    def resource_id(self):
        """Returns the resource id, used for making requests and checking responses.

        For base stations has the format [DEVICE-ID] and for other devices has
        the format [RESOURCE-TYPE]/[DEVICE-ID]
        """
        return self._device_id

    @property
    def resource_type(self):
        """Returns the type of resource this is.

        For now it's, `cameras`, `doorbells`, `lights` or `basestations`.
        """
        return None

    @property
    def serial_number(self):
        """Returns the device serial number."""
        return self._device_id

    @property
    def device_type(self):
        """Returns the Arlo reported device type.
\
        """
        return self._device_type

    @property
    def model_id(self):
        """Returns the model id."""
        return self._attrs.get("modelId", None)

    @property
    def hw_version(self):
        """Returns the hardware version."""
        return self._attrs.get("properties", {}).get("hwVersion", None)

    @property
    def timezone(self):
        """Returns the timezone."""
        return self._attrs.get("properties", {}).get("olsonTimeZone", None)

    @property
    def user_id(self):
        """Returns the user id."""
        return self._attrs.get("userId", None)

    @property
    def user_role(self):
        """Returns the user role."""
        return self._attrs.get("userRole", None)

    @property
    def xcloud_id(self):
        """Returns the device's xcloud id."""
        return self._load(XCLOUD_ID_KEY, "UNKNOWN")

    @property
    def web_id(self):
        """Return the device's web id."""
        return self.user_id + "_web"

    @property
    def unique_id(self):
        """Returns the device's unique id."""
        return self._unique_id

    @property
    def is_own_parent(self):
        """Returns True if device is its own parent.

        Can work from child or parent class.
        """
        return self._parent_id == self._device_id

    def attribute(self, attr, default=None):
        """Return the value of attribute attr.

        PyArlo stores its state in key/value pairs. This returns the value associated with the key.

        See PyArlo for a non-exhaustive list of attributes.

        :param attr: Attribute to look up.
        :type attr: str
        :param default: value to return if not found.
        :return: The value associated with attribute or `default` if not found.
        """
        value = self._load(attr, None)
        if value is None:
            value = self._attrs.get(attr, None)
        if value is None:
            value = self._attrs.get("properties", {}).get(attr, None)
        if value is None:
            value = default
        return value

    def add_attr_callback(self, attr, cb):
        """Add an callback to be triggered when an attribute changes.

        Used to register callbacks to track device activity. For example, get a notification whenever
        motion stop and starts.

        See PyArlo for a non-exhaustive list of attributes.

        :param attr: Attribute - eg `motionStarted` - to monitor.
        :type attr: str
        :param cb: Callback to run.
        """
        with self._lock:
            self._attr_cbs_.append((attr, cb))

    def has_capability(self, cap):
        """Is the device capable of performing activity cap:.

        Used to determine if devices can perform certain actions, like motion or audio detection.

        See attribute list against PyArlo.

        :param cap: Attribute - eg `motionStarted` - to check.
        :return: `True` it is, `False` it isn't.
        """
        if cap in (CONNECTION_KEY,):
            return True
        return False

    @property
    def state(self):
        """Returns a string describing the device's current state."""
        return "idle"

    @property
    def is_on(self):
        """Returns `True` if the device is on, `False` otherwise."""
        return True

    def turn_on(self):
        """Turn the device on."""
        pass

    def turn_off(self):
        """Turn the device off."""
        pass

    @property
    def is_unavailable(self):
        """Returns `True` if the device is unavailable, `False` otherwise.

        **Note:** Sorry about the double negative.
        """
        return self._load(CONNECTION_KEY, "unknown") == "unavailable"

    @property
    def battery_level(self):
        """Returns the current battery level."""
        return self._load(BATTERY_KEY, 100)

    @property
    def battery_tech(self):
        """Returns the current battery technology.

        Is it rechargable, wired...
        """
        return self._load(BATTERY_TECH_KEY, "None")

    @property
    def has_batteries(self):
        """Returns `True` if device has batteries installed, `False` otherwise."""
        return self.battery_tech != "None"

    @property
    def charger_type(self):
        """Returns how the device is recharging."""
        return self._load(CHARGER_KEY, "None")

    @property
    def has_charger(self):
        """Returns `True` if the charger is plugged in, `False` otherwise."""
        return self.charger_type != "None"

    @property
    def is_charging(self):
        """Returns `True` if the device is charging, `False` otherwise."""
        return self._load(CHARGING_KEY, "off").lower() == "on"

    @property
    def is_charger_only(self):
        """Returns `True` if the cahrger is plugged in with no batteries, `False` otherwise."""
        return self.battery_tech == "None" and self.has_charger

    @property
    def is_corded(self):
        """Returns `True` if the device is connected directly to a power outlet, `False` otherwise.

        The device can't have any battery option, it can't be using a charger, it has to run
        directly from a plug. ie, an original base station.
        """
        return not self.has_batteries and not self.has_charger

    @property
    def using_wifi(self):
        """Returns `True` if the device is connected to the wifi, `False` otherwise.

        This means connecting directly to your home wifi, not connecting to and Arlo basestation.
        """
        return self._attrs.get(CONNECTIVITY_KEY, {}).get("type", "").lower() == "wifi"

    @property
    def signal_strength(self):
        """Returns the WiFi signal strength (0-5)."""
        return self._load(SIGNAL_STR_KEY, 3)


class ArloChildDevice(ArloDevice):
    """Base class for all Arlo devices that attach to a base station."""

    def __init__(self, name, arlo, attrs):
        super().__init__(name, arlo, attrs)

        self._arlo.debug("parent is {}".format(self._parent_id))
        self._arlo.vdebug("resource is {}".format(self.resource_id))

    def _event_handler(self, resource, event):
        self._arlo.vdebug("{}: child got {} event **".format(self.name, resource))

        if resource.endswith("/states"):
            self._arlo.bg.run(self.base_station.update_mode)
            return

        # Pass event to lower level.
        super()._event_handler(resource, event)

    @property
    def resource_type(self):
        """Return the resource type this child device describes.

        Currently limited to `camera`, `doorbell` and `light`.
        """
        return "child"

    @property
    def resource_id(self):
        """Returns the child device resource id.

        Some devices - certain cameras - can provide other types.
        """
        return self.resource_type + "/" + self._device_id

    @property
    def parent_id(self):
        """Returns the parent device id.

        **Note:** Some devices - ArloBaby for example - are their own parents.
        """
        if self._parent_id is not None:
            return self._parent_id
        return self.device_id

    @property
    def base_station(self):
        """Returns the base station controlling this device.

        Some devices - ArloBaby for example - are their own parents. If we
        can't find a basestation, this returns the first one (if any exist).
        """
        # look for real parents
        for base in self._arlo.base_stations:
            if base.device_id == self.parent_id:
                return base

        # some cameras don't have base stations... it's its own base station...
        for base in self._arlo.base_stations:
            if base.device_id == self.device_id:
                return base

        # no idea!
        if len(self._arlo.base_stations) > 0:
            return self._arlo.base_stations[0]

        self._arlo.error("Could not find any base stations for device " + self._name)
        return None

    @property
    def is_unavailable(self):
        if not self.base_station:
            return True

        return (
            self.base_station.is_unavailable
            or self._load(CONNECTION_KEY, "unknown") == "unavailable"
        )

    @property
    def too_cold(self):
        """Returns `True` if the device too cold to operate, `False` otherwise."""
        return self._load(CONNECTION_KEY, "unknown") == "thermalShutdownCold"

    @property
    def state(self):
        if self.is_unavailable:
            return "unavailable"
        if not self.is_on:
            return "off"
        if self.too_cold:
            return "offline, too cold"
        return "idle"



import fnmatch
import pickle
import pprint
import threading


class ArloStorage(object):
    def __init__(self, arlo):
        self._arlo = arlo
        self._state_file = self._arlo.cfg.state_file
        self.db = {}
        self.lock = threading.Lock()
        self.load()

    def _ekey(self, key):
        return key if not isinstance(key, list) else "/".join(key)

    def _keys_matching(self, key):
        mkeys = []
        ekey = self._ekey(key)
        for mkey in self.db:
            if fnmatch.fnmatch(mkey, ekey):
                mkeys.append(mkey)
        return mkeys

    def load(self):
        if self._state_file is not None:
            try:
                with self.lock:
                    with open(self._state_file, "rb") as dump:
                        self.db = pickle.load(dump)
            except Exception:
                self._arlo.debug("file not read")

    def save(self):
        if self._state_file is not None:
            try:
                with self.lock:
                    with open(self._state_file, "wb") as dump:
                        pickle.dump(self.db, dump)
            except Exception:
                self._arlo.warning("file not written")

    def file_name(self):
        return self._state_file

    def get(self, key, default=None):
        with self.lock:
            ekey = self._ekey(key)
            return self.db.get(ekey, default)

    def get_matching(self, key, default=None):
        with self.lock:
            gets = []
            for mkey in self._keys_matching(key):
                gets.append((mkey, self.db.get(mkey, default)))
            return gets

    def keys_matching(self, key):
        with self.lock:
            return self._keys_matching(key)

    def set(self, key, value):
        ekey = self._ekey(key)
        output = "set:" + ekey + "=" + str(value)
        self._arlo.debug(output[:80])
        with self.lock:
            self.db[ekey] = value
            return value

    def unset(self, key):
        with self.lock:
            del self.db[self._ekey(key)]

    def clear(self):
        with self.lock:
            self.db = {}

    def dump(self):
        with self.lock:
            pprint.pprint(self.db)



import threading
import time
import traceback


class ArloBackgroundWorker(threading.Thread):
    def __init__(self, arlo):
        super().__init__()
        self._arlo = arlo
        self._id = 0
        self._lock = threading.Condition()
        self._queue = {}

    def _next_id(self):
        self._id += 1
        return str(self._id) + ":" + str(time.monotonic())

    def _run_next(self):

        # timeout in the future
        timeout = int(time.monotonic() + 60)

        # go by priority...
        for prio in sorted(self._queue.keys()):

            # jobs in particular priority
            for run_at, job_id in sorted(self._queue[prio].keys()):
                if run_at <= int(time.monotonic()):
                    job = self._queue[prio].pop((run_at, job_id))
                    self._lock.release()

                    # run it
                    try:
                        job["callback"](**job["args"])
                    except Exception as e:
                        self._arlo.error(
                            "job-error={}\n{}".format(
                                type(e).__name__, traceback.format_exc()
                            )
                        )

                    # reschedule?
                    self._lock.acquire()
                    run_every = job.get("run_every", None)
                    if run_every:
                        run_at += run_every
                        self._queue[prio][(run_at, job_id)] = job

                    # start going through list again
                    return None
                else:
                    if run_at < timeout:
                        timeout = run_at
                    break

        return timeout

    def run(self):

        with self._lock:
            while True:

                # loop till done
                timeout = None
                while timeout is None:
                    timeout = self._run_next()

                # wait or get going?
                now = time.monotonic()
                if now < timeout:
                    self._lock.wait(timeout - now)

    def queue_job(self, run_at, prio, job):
        run_at = int(run_at)
        with self._lock:
            job_id = self._next_id()
            if prio not in self._queue:
                self._queue[prio] = {}
            self._queue[prio][(run_at, job_id)] = job
            self._lock.notify()
        return job_id

    def stop_job(self, to_delete):
        with self._lock:
            for prio in self._queue.keys():
                for run_at, job_id in self._queue[prio].keys():
                    if job_id == to_delete:
                        # print( 'cancelling ' + str(job_id) )
                        del self._queue[prio][(run_at, job_id)]
                        return True
        return False


class ArloBackground:
    def __init__(self, arlo):
        self._worker = ArloBackgroundWorker(arlo)
        self._worker.name = "ArloBackgroundWorker"
        self._worker.daemon = True
        self._worker.start()
        arlo.debug("starting")

    def _run(self, bg_cb, prio, **kwargs):
        job = {"callback": bg_cb, "args": kwargs}
        return self._worker.queue_job(time.monotonic(), prio, job)

    def run_high(self, bg_cb, **kwargs):
        return self._run(bg_cb, 10, **kwargs)

    def run(self, bg_cb, **kwargs):
        return self._run(bg_cb, 40, **kwargs)

    def run_low(self, bg_cb, **kwargs):
        return self._run(bg_cb, 99, **kwargs)

    def _run_in(self, bg_cb, prio, seconds, **kwargs):
        job = {"callback": bg_cb, "args": kwargs}
        return self._worker.queue_job(time.monotonic() + seconds, prio, job)

    def run_high_in(self, bg_cb, seconds, **kwargs):
        return self._run_in(bg_cb, 10, seconds, **kwargs)

    def run_in(self, bg_cb, seconds, **kwargs):
        return self._run_in(bg_cb, 40, seconds, **kwargs)

    def run_low_in(self, bg_cb, seconds, **kwargs):
        return self._run_in(bg_cb, 99, seconds, **kwargs)

    def _run_every(self, bg_cb, prio, seconds, **kwargs):
        job = {"run_every": seconds, "callback": bg_cb, "args": kwargs}
        return self._worker.queue_job(time.monotonic() + seconds, prio, job)

    def run_high_every(self, bg_cb, seconds, **kwargs):
        return self._run_every(bg_cb, 10, seconds, **kwargs)

    def run_every(self, bg_cb, seconds, **kwargs):
        return self._run_every(bg_cb, 40, seconds, **kwargs)

    def run_low_every(self, bg_cb, seconds, **kwargs):
        return self._run_every(bg_cb, 99, seconds, **kwargs)

    def cancel(self, to_delete):
        if to_delete is not None:
            self._worker.stop_job(to_delete)



#!/usr/bin/env python
#

import base64
import io
import logging
import os
import pickle
import pprint
import sys

import click

from . import PyArlo
from .util import to_b64

logging.basicConfig(level=logging.ERROR,
                    format='%(asctime)s:%(name)s:%(levelname)s: %(message)s')
_LOGGER = logging.getLogger('pyaarlo')

BEGIN_PYAARLO_DUMP = "-----BEGIN PYAARLO DUMP-----"
END_PYAARLO_DUMP = "-----END PYAARLO DUMP-----"

PUBLIC_KEY = """-----BEGIN PUBLIC KEY-----
MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEA1oYXnbQPxREiVPUIRkgk
h+ehjxHnwz34NsjhjgN1oSKmHpf4cL4L/V4tMnj5NELEmLyTrzAZbeewUMwyiwXO
3l+cSjjoDKcPBSj4uxjWsq74Q5TLHGjOtkFwaqqxtvsVn3fGFWBO405xpvp7jPUc
BOvBQaUBUaR9Tbw5anMOzeavUwUTRp2rjtbWyj2P7PEp49Ixzw0w+RjIVrzzevAo
AD7SVb6U8P77fht4k9krbIFckC/ByY48HhmF+edh1GZAgLCHuf43tGg2upuH5wf+
AGv/Xlc+9ScTjEp37uPiCpHcB1ur83AFTjcceDIm+VDKF4zQrj88zmL7JqZy+Upx
UQIDAQAB
-----END PUBLIC KEY-----"""

opts = {
    "username": None,
    "password": None,

    "storage-dir": "./",
    "save-state": False,
    "dump-packets": False,
    "wait-for-initial-setup": True,

    "anonymize": False,
    "compact": False,
    "encrypt": False,
    "public-key": None,
    "private-key": "./rsa.private",
    "pass-phrase": None,
    "verbose": 0,
}

# where we store before encrypting or anonymizing
_out = None

# Arlo instance...
_arlo = None


def _debug(args):
    _LOGGER.debug("{}".format(args))


def _vdebug(args):
    if opts["verbose"] > 2:
        _debug(args)


def _info(args):
    _LOGGER.info("{}".format(args))


def _fatal(args):
    sys.exit("FATAL-ERROR:{}".format(args))


def _print_start():
    if opts['anonymize'] or opts['encrypt']:
        global _out
        _out = io.StringIO()


def _print(msg):
    if _out is None:
        print("{}".format(msg))
    else:
        _out.write(msg)
        _out.write("\n")


def _pprint(msg, obj):
    _print("{}\n{}".format(msg, pprint.pformat(obj, indent=1)))


def _print_end():
    if _out is not None:
        _out.seek(0, 0)
        out_text = _out.read()

        if opts['anonymize']:
            out_text = anonymize_from_string(out_text)
        if opts['encrypt']:
            print(BEGIN_PYAARLO_DUMP)
            out_text = encrypt_to_string(out_text)
        print(out_text)
        if opts['encrypt']:
            print(END_PYAARLO_DUMP)


def _casecmp(s1, s2):
    if s1 is None or s2 is None:
        return False
    return str(s1).lower() == str(s2).lower()


def encrypt_to_string(obj):
    from Crypto.Cipher import AES
    from Crypto.Random import get_random_bytes
    from Crypto.PublicKey import RSA
    from Crypto.Cipher import PKCS1_OAEP

    try:
        # pickle and resize object
        obj = pickle.dumps(obj)
        obj += b' ' * (16 - len(obj) % 16)

        # create key and encrypt pickled object with it
        key = get_random_bytes(16)
        aes_cipher = AES.new(key, AES.MODE_EAX)
        obj, tag = aes_cipher.encrypt_and_digest(obj)
        nonce = aes_cipher.nonce

        # encrypt key with public key
        if opts["public-key"] is None:
            rsa_cipher = RSA.importKey(PUBLIC_KEY)
        else:
            rsa_cipher = open(opts["public-key"], 'r').read()
            rsa_cipher = RSA.importKey(rsa_cipher)
        rsa_cipher = PKCS1_OAEP.new(rsa_cipher)
        key = rsa_cipher.encrypt(key)

        # create key/object dictionary, pickle and base64 encode
        key_obj = pickle.dumps({'k': key, 'n': nonce, 'o': obj, 't': tag})
        return base64.encodebytes(key_obj).decode().rstrip()
    except ValueError as err:
        _fatal("encrypt error: {}".format(err))
    except Exception:
        _fatal("unexpected encrypt error: {}".format(sys.exc_info()[0]))


def decrypt_from_string(key_obj):
    from Crypto.Cipher import AES
    from Crypto.PublicKey import RSA
    from Crypto.Cipher import PKCS1_OAEP

    try:
        # decode key/object dictionary then unpickle it
        key_obj = base64.b64decode(key_obj)
        key_obj = pickle.loads(key_obj)

        # import private key and decrypt nonce
        key = open(opts["private-key"], "r").read()
        rsa_cipher = RSA.importKey(key, passphrase=opts["pass-phrase"])
        rsa_cipher = PKCS1_OAEP.new(rsa_cipher)
        key = rsa_cipher.decrypt(key_obj["k"])

        # decrypt object and unpickle
        aes_cipher = AES.new(key, AES.MODE_EAX, nonce=key_obj['n'])
        obj = aes_cipher.decrypt_and_verify(key_obj['o'], key_obj['t'])
        obj = pickle.loads(obj)
        return obj
    except ValueError as err:
        _fatal("decrypt error {}".format(err))
    except Exception:
        _fatal("unexpected decrypt error: {}".format(sys.exc_info()[0]))


def anonymize_from_string(obj):
    # get device list
    keys = ['deviceId', 'uniqueId', 'userId', 'xCloudId']
    valuables = {}
    for device in _arlo._devices:
        for key in keys:
            value = device.get(key, None)
            if value and value not in valuables:
                valuables[value] = "X" * len(value)
        owner_id = device.get('owner', {}).get('ownerId', None)
        if owner_id:
            valuables[owner_id] = "X" * len(owner_id)
    if opts["username"] is not None:
        valuables[opts["username"]] = "USERNAME"
    if opts["password"] is not None:
        valuables[to_b64(opts["password"])] = "PASSWORD"

    anon = obj
    for valuable in valuables:
        anon = anon.replace(valuable, valuables[valuable])
    return anon


def login():
    _info("logging in")
    if opts["username"] is None or opts["password"] is None:
        _fatal("please supply a username and password")
    global _arlo
    _arlo = PyArlo(username=opts["username"], password=opts["password"],
                   storage_dir=opts["storage-dir"],
                   save_state=opts['save-state'],
                   wait_for_initial_setup=opts['wait-for-initial-setup'],
                   dump=opts['dump-packets']
                   )
    if _arlo is None:
        _fatal("unable to login to Arlo")
    return _arlo


def print_item(_name, item):
    if opts["compact"]:
        _print(" {};did={};mid={}/{};sno={}".format(item.name, item.device_id, item.model_id, item.hw_version,
                                                    item.serial_number))
    else:
        _print(" {}".format(item.name))
        _print("  device-id:{}".format(item.device_id))
        _print("  model-id:{}/{}".format(item.model_id, item.hw_version))
        _print("  serial-number:{}".format(item.serial_number))


def list_items(name, items):
    _print("{}:".format(name))
    if items is not None:
        for item in items:
            print_item(name, item)


# list [all|cameras|bases]
# describe device
# capture [encrpyted|to-file|wait]
#   all devices logs events
# encrypt [from-file|to-file]
#   encrypt 
# decrypt [from-file|to-file]
#   decrypt 

@click.group()
@click.option('-u', '--username', required=False,
              help="Arlo username")
@click.option('-p', '--password', required=False,
              help="Arlo password")
@click.option('-a', '--anonymize/--no-anonymize', default=False,
              help="Anonimize ids")
@click.option('-c', '--compact/--no-compact', default=False,
              help="Minimize lists")
@click.option('-e', '--encrypt/--no-encrypt', default=False,
              help="Where possible, encrypt output")
@click.option('-k', '--public-key', required=False,
              help="Public key for encryption")
@click.option('-K', '--private-key', required=False,
              help="Private key for decryption")
@click.option('-P', '--pass-phrase', required=False,
              help="Pass phrase for private key")
@click.option('-s', '--storage-dir',
              default="./", show_default='current dir',
              help="Where to store Arlo state and packet dump")
@click.option('-w', '--wait/--no-wait', default=True,
              help="Wait for all information to arrive starting up")
@click.option("-v", "--verbose", count=True,
              help="Be chatty. More is more chatty!")
def cli(username, password, anonymize, compact, encrypt, public_key, private_key, pass_phrase, storage_dir, wait,
        verbose):
    if username is not None:
        opts['username'] = username
    if password is not None:
        opts['password'] = password
    if anonymize is not None:
        opts['anonymize'] = anonymize
    if compact is not None:
        opts['compact'] = compact
    if encrypt is not None:
        opts['encrypt'] = encrypt
    if public_key is not None:
        opts['public-key'] = public_key
    if private_key is not None:
        opts['private-key'] = private_key
    if pass_phrase is not None:
        opts['pass-phrase'] = pass_phrase
    if storage_dir is not None:
        opts['storage-dir'] = storage_dir
    if wait is not None:
        opts['wait-for-initial-setup'] = wait
    if verbose is not None:
        opts['verbose'] = verbose
        if verbose == 0:
            _LOGGER.setLevel(logging.ERROR)
        if verbose == 1:
            _LOGGER.setLevel(logging.INFO)
        if verbose > 1:
            _LOGGER.setLevel(logging.DEBUG)


@cli.command()
@click.argument('item', default='all',
                type=click.Choice(['all', 'cameras', 'bases', 'lights', 'doorbells'], case_sensitive=False))
def dump(item):
    out = {}
    ar = login()

    if item == 'all':
        out = ar._devices

    _print_start()
    _pprint(item, out)
    _print_end()


@cli.command()
@click.argument('item', type=click.Choice(['all', 'cameras', 'bases', 'lights', 'doorbells'], case_sensitive=False))
def list(item):
    ar = login()
    _print_start()
    if item == "all" or item == "bases":
        list_items("bases", ar.base_stations)
    if item == "all" or item == "cameras":
        list_items("cameras", ar.cameras)
    if item == "all" or item == "lights":
        list_items("lights", ar.lights)
    if item == "all" or item == "doorbells":
        list_items("doorbells", ar.doorbells)
    _print_end()


@cli.command()
def encrypt():
    in_text = sys.stdin.read()
    enc_text = encrypt_to_string(in_text).rstrip()
    print("{}\n{}\n{}".format(BEGIN_PYAARLO_DUMP, enc_text, END_PYAARLO_DUMP))


@cli.command()
def decrypt():
    lines = ""
    save_lines = False
    for line in sys.stdin.readlines():
        if line.startswith(BEGIN_PYAARLO_DUMP):
            save_lines = True
        elif line.startswith(END_PYAARLO_DUMP):
            save_lines = False
        elif save_lines:
            lines += line
    if lines == "":
        _fatal('no encrypted input found')
    else:
        dec_text = decrypt_from_string(lines)
        print("{}".format(dec_text), end='')


@cli.command()
def anonymize():
    login()
    in_text = sys.stdin.read()
    print(anonymize_from_string(in_text))


@cli.command()
@click.option('-n', '--name', required=False,
              help='camera name')
@click.option('-d', '--device-id', required=False,
              help='camera device id')
@click.option('-f', '--start-ffmpeg/--no-start-ffmpeg', required=False, default=False,
              help='start ffmpeg for stream')
@click.argument('action', type=click.Choice(['start-stream', 'stop-stream', 'last-thumbnail'], case_sensitive=False))
def camera(name, device_id, start_ffmpeg, action):
    camera = None
    ar = login()
    for c in ar.cameras:
        if _casecmp(c.name, name) or _casecmp(c.device_id, device_id):
            camera = c
            break
    if camera is None:
        print('cannot find camera')
        return 0

    if action == 'start-stream':
        print('starting a stream')
        stream_url = camera.get_stream()
        if stream_url is None:
            print(' failed to start stream')
            return 0
        print("stream-url={}".format(stream_url))

        if start_ffmpeg:
            print('starting ffmpeg')
            os.system("mkdir video_dir")
            os.system("ffmpeg -i '{}' ".format(stream_url) +
                      "-fflags flush_packets -max_delay 2 -flags -global_header " +
                      "-hls_time 2 -hls_list_size 3 -vcodec copy -y video_dir/video.m3u8")

    elif action == 'stop-stream':
        pass

    elif action == 'last-thumbnail':
        last_thumbnail = camera.last_thumbnail
        if last_thumbnail:
            print("last-thumbnail={}".format(last_thumbnail))
        else:
            print(' error getting thumbnail')


def main_func():
    cli()


if __name__ == '__main__':
    cli()



from .constant import (
    DEFAULT_AUTH_HOST,
    DEFAULT_HOST,
    PRELOAD_DAYS,
    TFA_CONSOLE_SOURCE,
    TFA_DEFAULT_HOST,
    TFA_DELAY,
    TFA_EMAIL_TYPE,
    TFA_RETRIES,
)


class ArloCfg(object):
    """Helper class to get at Arlo configuration options.

    I got sick of adding in variables each time the config changed so I moved it all here. Config
    is passed in a kwarg and parsed out by the property methods.

    """

    def __init__(self, arlo, **kwargs):
        """The constructor.

        Args:
            kwargs (kwargs): Configuration options.

        """
        self._arlo = arlo
        self._kw = kwargs
        self._arlo.debug("Cfg started")

    @property
    def storage_dir(self):
        return self._kw.get("storage_dir", "/config/.aarlo")

    @property
    def name(self):
        return self._kw.get("name", "aarlo")

    @property
    def username(self):
        return self._kw.get("username", "unknown")

    @property
    def password(self):
        return self._kw.get("password", "unknown")

    @property
    def host(self):
        return self._kw.get("host", DEFAULT_HOST)

    @property
    def auth_host(self):
        return self._kw.get("auth_host", DEFAULT_AUTH_HOST)

    @property
    def dump(self):
        return self._kw.get("dump", False)

    @property
    def max_days(self):
        return self._kw.get("max_days", 365)

    @property
    def db_motion_time(self):
        return self._kw.get("db_motion_time", 30)

    @property
    def db_ding_time(self):
        return self._kw.get("db_ding_time", 10)

    @property
    def request_timeout(self):
        return self._kw.get("request_timeout", 60)

    @property
    def stream_timeout(self):
        return self._kw.get("stream_timeout", 0)

    @property
    def recent_time(self):
        return self._kw.get("recent_time", 600)

    @property
    def last_format(self):
        return self._kw.get("last_format", "%m-%d %H:%M")

    @property
    def no_media_upload(self):
        return self._kw.get("no_media_upload", False)

    @property
    def media_retry(self):
        retries = self._kw.get("media_retry", [])
        if not retries and self.no_media_upload:
            retries = [0, 5, 10]
        return retries

    @property
    def snapshot_checks(self):
        checks = self._kw.get("snapshot_checks", [])
        if not checks:
            return [1, 5]
        return checks

    @property
    def user_agent(self):
        return self._kw.get("user_agent", "arlo")

    @property
    def mode_api(self):
        return self._kw.get("mode_api", "auto")

    @property
    def refresh_devices_every(self):
        return self._kw.get("refresh_devices_every", 0) * 60 * 60

    @property
    def refresh_modes_every(self):
        return self._kw.get("refresh_modes_every", 0) * 60

    @property
    def http_connections(self):
        return self._kw.get("http_connections", 20)

    @property
    def http_max_size(self):
        return self._kw.get("http_maz_size", 10)

    @property
    def reconnect_every(self):
        return self._kw.get("reconnect_every", 0) * 60

    @property
    def snapshot_timeout(self):
        return self._kw.get("snapshot_timeout", 60)

    @property
    def verbose(self):
        return self._kw.get("verbose_debug", False)

    @property
    def hide_deprecated_services(self):
        return self._kw.get("hide_deprecated_services", False)

    @property
    def tfa_source(self):
        return self._kw.get("tfa_source", TFA_CONSOLE_SOURCE)

    @property
    def tfa_type(self):
        return self._kw.get("tfa_type", TFA_EMAIL_TYPE).lower()

    @property
    def tfa_delay(self):
        return self._kw.get("tfa_delay", TFA_DELAY)

    @property
    def tfa_retries(self):
        return self._kw.get("tfa_retries", TFA_RETRIES)

    @property
    def tfa_timeout(self):
        return self._kw.get("tfa_timeout", 3)

    @property
    def tfa_total_timeout(self):
        return self._kw.get("tfa_total_timeout", 60)

    @property
    def tfa_host(self):
        return self._kw.get("tfa_host", TFA_DEFAULT_HOST)

    @property
    def tfa_username(self):
        u = self._kw.get("tfa_username", None)
        if u is None:
            u = self.username
        return u

    @property
    def tfa_password(self):
        p = self._kw.get("tfa_password", None)
        if p is None:
            p = self.password
        return p

    @property
    def wait_for_initial_setup(self):
        return self._kw.get("wait_for_initial_setup", True)

    @property
    def save_state(self):
        return self._kw.get("save_state", True)

    @property
    def state_file(self):
        if self.save_state:
            return self.storage_dir + "/" + self.name + ".pickle"
        return None

    @property
    def dump_file(self):
        if self.dump:
            return self.storage_dir + "/" + "packets.dump"
        return None

    @property
    def library_days(self):
        return self._kw.get("library_days", PRELOAD_DAYS)

    @property
    def synchronous_mode(self):
        return self._kw.get("synchronous_mode", False)

    @property
    def user_stream_delay(self):
        return self._kw.get("user_stream_delay", 1)

    @property
    def serial_ids(self):
        return self._kw.get("serial_ids", False)

    @property
    def stream_snapshot(self):
        return self._kw.get("stream_snapshot", False)

    @property
    def stream_snapshot_stop(self):
        return self._kw.get("stream_snapshot_stop", 10)

    @property
    def save_updates_to(self):
        return self._kw.get("save_updates_to", "")

    @property
    def save_media_to(self):
        return self._kw.get("save_media_to", "")

    @property
    def no_unicode_squash(self):
        return self._kw.get("no_unicode_squash", True)



import pprint

from .constant import BATTERY_KEY, BRIGHTNESS_KEY, LAMP_STATE_KEY, MOTION_DETECTED_KEY
from .device import ArloChildDevice


class ArloLight(ArloChildDevice):
    def __init__(self, name, arlo, attrs):
        """An Arlo Light.

        :param name: name of light
        :param arlo: controlling arlo instance
        :param attrs: initial attributes give by Arlo
        """
        super().__init__(name, arlo, attrs)

    @property
    def resource_type(self):
        return "lights"

    def _event_handler(self, resource, event):
        self._arlo.debug(self.name + " LIGHT got one " + resource)

        # pass on to lower layer
        super()._event_handler(resource, event)

    @property
    def is_on(self):
        return self._load(LAMP_STATE_KEY, "off") == "on"

    def turn_on(self, brightness=None, rgb=None):
        """Turn the light on.

        :param brightness: how bright to make the light
        :param rgb: what color to make the light
        """
        properties = {LAMP_STATE_KEY: "on"}
        if brightness is not None:
            properties[BRIGHTNESS_KEY] = brightness
        if rgb is not None:
            # properties["single"] = rgb_to_hex(rgb)
            pass

        self._arlo.debug("{} sending {}".format(self._name, pprint.pformat(properties)))
        self._arlo.be.notify(
            base=self.base_station,
            body={
                "action": "set",
                "properties": properties,
                "publishResponse": True,
                "resource": self.resource_id,
            },
        )
        return True

    def turn_off(self):
        """Turn the light off."""
        self._arlo.be.notify(
            base=self.base_station,
            body={
                "action": "set",
                "properties": {LAMP_STATE_KEY: "off"},
                "publishResponse": True,
                "resource": self.resource_id,
            },
        )
        return True

    def set_brightness(self, brightness):
        """Set the light brightness.

        :param brightness: brightness to use (0-255)
        """
        self._arlo.be.notify(
            base=self.base_station,
            body={
                "action": "set",
                "properties": {BRIGHTNESS_KEY: brightness},
                "publishResponse": True,
                "resource": self.resource_id,
            },
        )
        return True

    def has_capability(self, cap):
        if cap in (MOTION_DETECTED_KEY, BATTERY_KEY):
            return True
        return super().has_capability(cap)
